/*
 * Copyright (C) 2008-2009, 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "exif.h"

#include <iostream>
#include "../shared/lib_conversions.h"
#include "keyword.h"

#if (EXIV2_FOUND == 1)
# include <iomanip>
#endif

std::string clean_exiv2(std::string in)
{
  std::string::size_type pos = in.find('\0');
  if (pos != std::string::npos)
  {
    std::cerr << "there's something wrong with exiv2 data...\n";
    return(in.substr(0, pos));
  }
  else
    return(in);
}

ExifInfo::ExifInfo(std::string filename)
{
  if (strU(filename).find("JPG")  == filename.length()-3 ||
      strU(filename).find("JPEG") == filename.length()-4)
  {
#ifdef USE_FLEXIF   
    exif = new Fl_EXIF_Data(filename.c_str());
#endif
#ifdef USE_LIBEXIF
    ed = exif_data_new_from_file(filename.c_str());
#endif
#if (EXIV2_FOUND == 1)
    try 
    {
      image = Exiv2::ImageFactory::open(filename.c_str());
      image->readMetadata();
    }
    catch (Exiv2::AnyError& e) 
    {
      std::cerr << "Caught Exiv2 exception '" << e << "'\n";
    }
#endif
  }
  else
  {
#ifdef USE_FLEXIF   
    exif = (Fl_EXIF_Data*)0;
#endif
#ifdef USE_LIBEXIF
    ed = (ExifData*)0;
#endif
  }
}

ExifInfo::~ExifInfo()
{
#ifdef USE_FLEXIF
  if (exif != (Fl_EXIF_Data*)0)
    delete exif;
#endif
#ifdef USE_LIBEXIF
  if (ed != (ExifData*)0)
    exif_data_free(ed);
#endif
}

bool ExifInfo::hasEXIF()
{
#ifdef USE_FLEXIF   
  return(exif != (Fl_EXIF_Data*)0 && exif->exif_count());
#endif
#ifdef USE_LIBEXIF
  return(ed != (ExifData*)0 && ed->size != 0);
#endif
#if (EXIV2_FOUND == 1)
  bool fHas = false;
  try
  {
    if (image.get() != 0)
    {
      Exiv2::ExifData& exifData = image->exifData();
      fHas = !exifData.empty();
    }    
  }
  catch (Exiv2::AnyError& e) 
  {
  }    
  return(fHas);
#endif
}

std::string ExifInfo::getInfo()
{
  if (!hasEXIF())
    return("[no exif info]");
  
  std::string tmp = "";
  
#ifdef USE_FLEXIF   
  tmp += getExifString(exif, Fl_EXIF_Data::TAG_MAKE);
  tmp += "\n";
  tmp += getExifString(exif, Fl_EXIF_Data::TAG_MODEL);
  tmp += "\n";
  
  double val = exif->get_rational(Fl_EXIF_Data::TAG_EXPOSURE_TIME);
  if (val < 1.0)
  {
    tmp += "1/";
    tmp += ftoStr(1.0/val, 3, 0);
  }
  else
  {
    tmp += ftoStr(val, 2, 1);
  }
  tmp += " s\n";
  tmp += getDateTime();
#endif
  
#ifdef USE_LIBEXIF   
  ExifTag tags[] =
  {
      EXIF_TAG_MAKE,
      EXIF_TAG_MODEL,
      
      EXIF_TAG_FNUMBER,
      EXIF_TAG_EXPOSURE_TIME,
      EXIF_TAG_EXPOSURE_BIAS_VALUE,
      
      EXIF_TAG_FLASH,
      EXIF_TAG_FLASH_ENERGY,
            
      EXIF_TAG_WHITE_BALANCE,
      
      EXIF_TAG_FOCAL_LENGTH,
      EXIF_TAG_FOCAL_LENGTH_IN_35MM_FILM,
      
      EXIF_TAG_EXPOSURE_INDEX,
      EXIF_TAG_EXPOSURE_MODE,
      EXIF_TAG_EXPOSURE_PROGRAM,
      
      //      EXIF_TAG_JPEG_PROC,
      
      EXIF_TAG_DATE_TIME_ORIGINAL,
      EXIF_TAG_DATE_TIME_DIGITIZED,
      EXIF_TAG_DATE_TIME,      
      
      EXIF_TAG_SOFTWARE,
      
      EXIF_TAG_ORIENTATION,
      
      EXIF_TAG_WHITE_POINT,
      EXIF_TAG_SPECTRAL_SENSITIVITY,
      EXIF_TAG_ISO_SPEED_RATINGS,
      EXIF_TAG_SHUTTER_SPEED_VALUE,
            
      EXIF_TAG_SCENE_TYPE,
      EXIF_TAG_SCENE_CAPTURE_TYPE,
      
      EXIF_TAG_SENSING_METHOD,
      EXIF_TAG_LIGHT_SOURCE,
      EXIF_TAG_APERTURE_VALUE,
      EXIF_TAG_BRIGHTNESS_VALUE,
      EXIF_TAG_METERING_MODE,
      
      //      EXIF_TAG_MAKER_NOTE,
      //      EXIF_TAG_SUB_SEC_TIME,
      //      EXIF_TAG_SUB_SEC_TIME_ORIGINAL,
      //      EXIF_TAG_SUB_SEC_TIME_DIGITIZED,
      // 
      EXIF_TAG_SUBJECT_AREA,
      EXIF_TAG_SUBJECT_LOCATION,
      EXIF_TAG_SUBJECT_DISTANCE,
      EXIF_TAG_SUBJECT_DISTANCE_RANGE,
      
      EXIF_TAG_CUSTOM_RENDERED,
      
      EXIF_TAG_GAIN_CONTROL,
      EXIF_TAG_CONTRAST,
      EXIF_TAG_SATURATION,
      EXIF_TAG_SHARPNESS,
      
      EXIF_TAG_DEVICE_SETTING_DESCRIPTION,
      
      EXIF_TAG_DIGITAL_ZOOM_RATIO
  };
  
  for (unsigned int n=0; n<sizeof(tags)/sizeof(ExifTag); n++)
  {
    // This line outputs something for the help file fa_exif.htm.
    //      std::cout << "<tr><td>" << exif_tag_get_name(tags[n]) << "</td>\n    <td>" << exif_tag_get_description(tags[n]) << "</td></tr>\n";
              
    tmp += getExifString(ed, tags[n]);
    tmp += "\n";
  }  
#endif
  
#if (EXIV2_FOUND == 1)
  try
  {
    if (image.get() != 0)
    {
      Exiv2::ExifData& exifData = image->exifData();
      
      tmp += get(exifData, "Exif.Image.Make") + "\n";
      tmp += get(exifData, "Exif.Image.Model") + "\n";

      tmp += get(exifData, "Exif.Photo.FNumber") + "\n";
      tmp += get(exifData, "Exif.Photo.ExposureTime") + "\n";
      tmp += get(exifData, "Exif.Photo.ExposureBiasValue") + "\n";
      tmp += get(exifData, "Exif.Photo.ISOSpeedRatings") + "\n";
      tmp += get(exifData, "Exif.Photo.Flash") + "\n";
      tmp += get(exifData, "Exif.Photo.WhiteBalance") + "\n";

      tmp += get(exifData, "Exif.Photo.FocalLength") + "\n";
      tmp += get(exifData, "Exif.Photo.FocalLengthIn35mmFilm") + "\n";

      tmp += get(exifData, "Exif.Photo.ExposureMode") + "\n";

      tmp += get(exifData, "Exif.Photo.DateTimeOriginal") + "\n";
      tmp += get(exifData, "Exif.Photo.DateTimeDigitized") + "\n";
      tmp += get(exifData, "Exif.Image.DateTime") + "\n";

      tmp += get(exifData, "Exif.Image.Software") + "\n";
      tmp += get(exifData, "Exif.Photo.MaxApertureValue") + "\n";
      tmp += get(exifData, "Exif.Photo.MeteringMode") + "\n";
      tmp += get(exifData, "Exif.Photo.CustomRendered") + "\n";
      tmp += get(exifData, "Exif.Photo.DigitalZoomRatio") + "\n";
      tmp += get(exifData, "Exif.Photo.SceneCaptureType") + "\n";
      tmp += get(exifData, "Exif.Photo.Contrast") + "\n";
      tmp += get(exifData, "Exif.Photo.Saturation") + "\n";
      tmp += get(exifData, "Exif.Photo.Sharpness") + "\n";
      tmp += get(exifData, "Exif.Photo.SubjectDistanceRange") + "\n";
    }
  }
  catch (Exiv2::AnyError& e) 
  {
  }    
#endif
  
  return(tmp);
}

std::string ExifInfo::getDateTime()
{
  if (!hasEXIF())
    return("");
  
  std::string tmp;
  
#ifdef USE_FLEXIF 
  const char* date_time = exif->get_ascii(Fl_EXIF_Data::TAG_DATE_TIME);
  
  if (date_time != (const char*)0)
  {
    tmp = date_time;
  }
#endif
#ifdef USE_LIBEXIF
  ExifEntry* entry;
  char       descr[100];
  
  /**
   * Die DATE_TIME-Tags geben zwar bei einigen Kameras alle die gleiche Zeit
   * (nämlich die der Aufnahme) an, bei anderen aber ist EXIF_TAG_DATE_TIME
   * eine spätere Zeit (runterladen der Fotos?).
   */
  entry = exif_data_get_entry(ed, EXIF_TAG_DATE_TIME_ORIGINAL);
  if (entry == (ExifEntry*)0)
    entry = exif_data_get_entry(ed, EXIF_TAG_DATE_TIME_DIGITIZED);
  if (entry == (ExifEntry*)0)
    entry = exif_data_get_entry(ed, EXIF_TAG_DATE_TIME);
  
  if (entry != (ExifEntry*)0)
  {
    exif_entry_get_value(entry, descr, 99);
    tmp = descr;
    //    exif_entry_free(entry);
  }
#endif
#if (EXIV2_FOUND == 1)
  try
  {
    if (image.get() != 0)
    {
      Exiv2::ExifData& exifData = image->exifData();
      
      tmp = get(exifData, "Exif.Photo.DateTimeOriginal", false);
      if (tmp.length() != 19)
      {
        tmp = get(exifData, "Exif.Photo.DateTimeDigitized", false);
        if (tmp.length() != 19)
          tmp = get(exifData, "Exif.Image.DateTime", false);
      }
    }
  }
  catch (Exiv2::AnyError& e) 
  {
  }    
#endif
  
  if (tmp.length() == 19)
    return(tmp);
  else
    return("");
}


#ifdef USE_FLEXIF 
std::string ExifInfo::getExifString(Fl_EXIF_Data* exif, 
                                    unsigned      tag)
{
  const char* date_time = exif->get_ascii(tag);
  
  if (date_time != (const char*)0)
    return(date_time);
  else
    return("");
}
#endif

#ifdef USE_LIBEXIF
std::string ExifInfo::getExifString(ExifData*   ed,
                                    ExifTag     tag)
{
  ExifEntry*  entry;
  char        descr[100];
  std::string tmp;
  
  
  entry = exif_data_get_entry(ed, tag);
  if (entry != (ExifEntry*)0)
  {  
    //    std::cout << tmp << " " << (entry->data) << " " << (entry->size) << " " << (entry->parent) << " " << (entry->priv) << "\n";
    tmp = exif_tag_get_name(tag);
    tmp += ": ";
    exif_entry_get_value(entry, descr, 99);
    tmp += descr;
  }
  else
  {
    tmp = '-';
  }
  
  return(tmp);
}
#endif

#if (EXIV2_FOUND == 1)
std::string ExifInfo::get(const Exiv2::ExifData& exifData,
                          const std::string& key, bool fLabel)
{
  Exiv2::ExifKey ek(key);
  Exiv2::ExifData::const_iterator md = exifData.findKey(ek);
  if (md != exifData.end()) 
  {
    if (fLabel)
      return(md->tagLabel() + ": " + md->toString());
    else
      return(md->toString());
  }
  else
    return("");
}

std::string ExifInfo::get(const std::string& key)
{
  std::string tmp = "";
  
  try
  {
    if (image.get() != 0)
    {
      Exiv2::ExifData& exifData = image->exifData();      
      tmp = get(exifData, key, false);
    }
  }
  catch (Exiv2::AnyError& e) 
  {
  }    
  
  return(tmp);  
}


void ExifInfo::ReadKeywords(Exiv2::Image::AutoPtr& image, std::set<std::string>& kwset)
{
  try
  {
    if (image.get() != 0)
    {
# if 1
      Exiv2::XmpData& xmpData = image->xmpData();
      if (!xmpData.empty())
      {
        for (Exiv2::XmpData::const_iterator md = xmpData.begin();
             md != xmpData.end(); ++md) 
        {
#  if 0
          std::cout << std::setfill(' ') << std::left
            << std::setw(44)
              << md->key() << " "
            << std::setw(9) << std::setfill(' ') << std::left
            << md->typeName() << " "
            << std::dec << std::setw(3)
              << std::setfill(' ') << std::right
            << md->count() << "  "
            << std::dec << md->value()
           << std::endl;
#  endif
          if (md->key().compare("Xmp.dc.subject") == 0)
          {
            for (int n=0; n < md->count(); n++)
              kwset.insert(clean_exiv2(md->toString(n)));
          }
        }
      }      
      Exiv2::XmpParser::terminate();
#endif
      
      //
      Exiv2::IptcData &iptcData = image->iptcData();
      if (!iptcData.empty())
      {
        std::string loc_country = "_";
        std::string loc_state   = "_";
        std::string loc_city    = "_";
        std::string loc_sub     = "";
        
        Exiv2::IptcData::iterator end = iptcData.end();      
        for (Exiv2::IptcData::iterator md = iptcData.begin(); md != end; ++md) 
        {
          /*
          std::cout << std::setw(44) << std::setfill(' ') << std::left
            << md->key() << " "
            << "0x" << std::setw(4) << std::setfill('0') << std::right
            << std::hex << md->tag() << " "
            << std::setw(9) << std::setfill(' ') << std::left
            << md->typeName() << " "
            << std::dec << std::setw(3)
              << std::setfill(' ') << std::right
            << md->count() << "  "
            << std::dec << md->value()
              << std::endl;
           */
          
          if (md->key().compare("Iptc.Application2.Keywords")     == 0 ||
              md->key().compare("Iptc.Application2.SuppCategory") == 0)
            kwset.insert(clean_exiv2(md->toString()));
          
          if (md->key().compare("Iptc.Application2.CountryName") == 0)
            loc_country = Keywords::Keyword(clean_exiv2(md->toString()));
          if (md->key().compare("Iptc.Application2.ProvinceState") == 0)
            loc_state = Keywords::Keyword(clean_exiv2(md->toString()));
          if (md->key().compare("Iptc.Application2.City") == 0)
            loc_city = Keywords::Keyword(clean_exiv2(md->toString()));
          if (md->key().compare("Iptc.Application2.SubLocation") == 0)
            loc_sub = Keywords::Keyword(clean_exiv2(md->toString()));
        }
        
        if (loc_country.length() > 1 || loc_state.length() > 1 || loc_city.length() > 1 || loc_sub.length())
        {
          kwset.insert("location." + loc_country
                       += "." + loc_state
                       += "." + loc_city
                       += "." + loc_sub);
        }
      }
    }
  }
  catch (Exiv2::AnyError& e) 
  {
  }    
}
#endif

void ExifInfo::ReadKeywords(std::set<std::string>& kwset)
{
#if (EXIV2_FOUND == 1)
  ReadKeywords(image, kwset);
#endif
}

std::string ExifInfo::ReadComment(std::string old)
{
  std::string comment;
  std::string comment_jpeg;
  std::string comment_iptc;
  
#if (EXIV2_FOUND == 1)
  try
  {
    if (image.get() != 0)
    {
      std::string commentstr;
      
      // read jpeg comment
      {
        commentstr = image->comment();
        if (commentstr.length() && 
            commentstr.compare("Created with The GIMP") != 0 &&
            old.find(commentstr) == std::string::npos)
          comment_jpeg = commentstr + "\n";
      }
      
      Exiv2::IptcData &iptcData = image->iptcData();
      if (!iptcData.empty())
      {
        Exiv2::IptcData::iterator end = iptcData.end();
        for (Exiv2::IptcData::iterator md = iptcData.begin(); md != end; ++md) 
        {
          /*
          std::cout << std::setw(44) << std::setfill(' ') << std::left
            << md->key() << " "
            << "0x" << std::setw(4) << std::setfill('0') << std::right
            << std::hex << md->tag() << " "
            << std::setw(9) << std::setfill(' ') << std::left
            << md->typeName() << " "
            << std::dec << std::setw(3)
              << std::setfill(' ') << std::right
            << md->count() << "  "
            << std::dec << md->value()
           << std::endl;
           */
          
          if ((md->key().compare("Iptc.Application2.Headline") == 0 ||
               md->key().compare("Iptc.Application2.ObjectName") == 0) &&
              old.find(clean_exiv2(md->toString())) == std::string::npos
              )
            comment_iptc += std::string(clean_exiv2(md->toString())) + "\n";
          
          if (md->key().compare("Iptc.Application2.Caption") == 0 &&
              old.find(clean_exiv2(md->toString())) == std::string::npos)
            comment_iptc += std::string(clean_exiv2(md->toString())) + "\n";
        }
      }
      
      if (comment_jpeg.length())
      {
        comment += old + comment_jpeg;
        
        if (comment_iptc.length() && comment_jpeg.compare(comment_iptc) != 0)
          comment += comment_iptc;
      }
      else if (comment_iptc.length())
        comment += old + comment_iptc;

      return(comment);
    }
  }
  catch (Exiv2::AnyError& e) 
  {
    return(old);
  }    
#endif
  return(old);
}

std::string ExifInfo::Get_Make()
{
# ifdef EXIV2_FOUND
  return get("Exif.Image.Make");
# else
  return "";
# endif
}

std::string ExifInfo::Get_Model()
{
# ifdef EXIV2_FOUND
  return get("Exif.Image.Model");
# else
  return "";
# endif
}

int ExifInfo::Get_ISO()
{
# ifdef EXIV2_FOUND
  std::string strISO = get("Exif.Photo.ISOSpeedRatings");
  if (strISO.length())
    return atoi(strISO.c_str());
  else   
    return 0;
# else
  return 0;
# endif
}

int ExifInfo::Get_ExposureTime_us()
{
# ifdef EXIV2_FOUND
  std::string strISO = get("Exif.Photo.ExposureTime");
  if (strISO.length())
  {
    std::string::size_type p = strISO.find("/");
    if (p == std::string::npos)
      return 1000000*atoi(strISO.c_str());
    else
      return (1000000*atoi(strISO.substr(0, p).c_str()))/atoi(strISO.substr(p+1).c_str());
  }
  else
    return 0;
# else
  return 0;
# endif
}

int ExifInfo::Get_FNumber_m10()
{
# ifdef EXIV2_FOUND
  std::string strISO = get("Exif.Photo.FNumber");
  if (strISO.length())
  {
    std::string::size_type p = strISO.find("/");
    if (p == std::string::npos)
      return 10*atoi(strISO.c_str());
    else
      return (10*atoi(strISO.substr(0, p).c_str()))/atoi(strISO.substr(p+1).c_str());
  }
  else   
    return 0;
# else
  return 0;
# endif
}

int ExifInfo::Get_focalLen()
{
# ifdef EXIV2_FOUND
  std::string strFL = get("Exif.Photo.FocalLength");
  if (strFL.length())
  {
    std::string::size_type p = strFL.find("/");
    if (p == std::string::npos)
      return atoi(strFL.c_str());
    else
      return (atoi(strFL.substr(0, p).c_str()))/atoi(strFL.substr(p+1).c_str());
  }
  else   
    return 0;
# else
  return 0;
# endif
}
