/*
 * Copyright (C) 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "group.h"

#include <iostream>
#include <sys/stat.h>
#include <unistd.h>
#include "xmldefs.h"
#include "../shared/lib_conversions.h"
#include "../shared/scandir.h"
#include "../shared/filetools.h"
#include "../shared/hirestime.h"
#include "treeview.h"
#include "searchprog.h"
#include "../shared/utf8.h"

ItGroup::ItGroup()
{
  strID       = "";
  nRefCnt     = 0;
  pFirstChild = 0;
  pNext       = 0;  
  pParent     = 0;
  pPrev       = 0;  
  fExpanded   = false;
}

ItGroup::ItGroup(ItGroup* pPrev, ItGroup* pParent)
{
  strID       = "";
  
  nRefCnt     = 0;
  pFirstChild = 0;
  pNext       = 0;
  
  this->pParent = pParent;
  this->pPrev   = pPrev;
  if (pPrev != 0)
    pPrev->pNext = this;
  if (pParent != 0 && pPrev == 0)
    pParent->pFirstChild = this;
  fExpanded = (pParent == 0);    
}

ItGroup* ItGroup::Factory(SimpleXMLTransfer* xIn)
{
  if (xIn->indexOfAttribute(XMLNAME_SAVE_SEARCH) >= 0)
    return new ItGSearch(xIn);
  else if (xIn->indexOfAttribute(XMLNAME_SAVE_LINK) >= 0)
    return new ItGLinkRef(xIn);
  else
    return new ItGData(xIn);  
}

int ItGroup::SetChildren(SimpleXMLTransfer* xIn, std::map<std::string, ItPic*>& mapPics, std::map<std::string, ItGroup*>& mapGroups)
{
  int retCode = 0;
  std::string strChildren = xIn->getContentString();  
  std::string::size_type posA, posB;
  std::string strLink;
  posA = 0;
  pFirstChild = 0;
  ItGroup* pLastGroup = 0;  
  while (strChildren.length() > posA)
  {
    // Jetzt zeigt posA auf den Anfang eines IDs
    posB = posA + 1;
    while (strChildren.length() > posB && strChildren[posB] != '\n')
      posB++;
    // posB steht hinter dem ID
    strLink = strChildren.substr(posA, posB-posA);
    
    // 
    if (islower(strLink[0]) == 0)
    {
      std::map<std::string, ItGroup*>::iterator itG = mapGroups.find(strLink);
      if (itG == mapGroups.end())
      {
        std::cerr << __LINE__ << ", " << GetName() << ", " << GetID() << ", Link target not found: *" << strLink << "*\n";
        retCode |= 1;
      }
      else
      {
        if (pFirstChild)
        {
          pLastGroup->pNext = itG->second;
          itG->second->pPrev = pLastGroup;          
        }
        else
        {
          pFirstChild = itG->second;
          pFirstChild->pPrev = 0;
        }
        pLastGroup = itG->second;
        
        itG->second->SetParent(this);
        itG->second->pNext = 0;
      }
    }
    else
    {
      std::map<std::string, ItPic*>::iterator itG = mapPics.find(strLink);
      if (itG == mapPics.end())
      {
        std::cerr << __LINE__ << ", " << GetName() << ", " << GetID() << ", Link target not found: *" << strLink << "*\n";
        retCode |= 1;
      }
      else
      {
        vPics.push_back(new ItPicInGroup(itG->second));
        itG->second->IncRefCnt();
      }
    }
    
    posA = posB + 1;
  }
  
  retCode |= SetLink(xIn, mapGroups);
  
  return retCode;
}

int ItGroup::treeDepth()                                           
{
  int       nNum = 0;
  ItGroup* item = pParent;
  
  while (item != (ItGroup*)0)
  {
    item = item->pParent;
    nNum++;
  }
  
  return(nNum);
}

unsigned int ItGroup::getMinMaxDateTime(JWWDateTime& min,
                                        JWWDateTime& max,
                                        bool         fRecursive)          
{
  unsigned int  nNonGroup = 0;

  if (fRecursive)
  {
    ItGroup* ptr = pFirstChild;
    while (ptr != (ItGroup*)0)
    {
      ptr->getMinMaxDateTime(min, max, fRecursive);
      ptr = ptr->pNext;
    }
  }
  
  nNonGroup += vPics.size();
  
  for (unsigned int n=0; n<vPics.size(); n++)
  {
    ItPic* ptr = vPics[n]->pPic;
    if (ptr->hasDate())
    {
      JWWDateTime dt = ptr->getDT();
      
      if (!min.isDateSet())
        min = dt;
      else if (min.compare(&dt) < 0)
        min = dt;
      
      if (!max.isDateSet())
        max = dt;
      else if (max.compare(&dt) > 0)
        max = dt;
    }
  }
  
  return(nNonGroup);
}

ItPicInGroup* ItGroup::GetPicNext(ItPic* pPic)
{
  for (unsigned int n=0; n<vPics.size(); n++)
    if (vPics[n]->pPic == pPic)
    {
      if (n < vPics.size()-1)
        return vPics[n+1];
      else
        return 0;
    }
  
  return 0;
}

ItPicInGroup* ItGroup::GetPicPrev(ItPic* pPic)
{
  for (unsigned int n=0; n<vPics.size(); n++)
    if (vPics[n]->pPic == pPic)
    {
      if (n > 0)
        return vPics[n-1];
      else
        return 0;
    }
  
  return 0;
}

void ItGroup::GetMarkedPics(std::vector<ItPicInGroup*>& vMarkedPics)
{
  for (unsigned int n=0; n<vPics.size(); n++)
    if (vPics[n]->fMarked)
      vMarkedPics.push_back(vPics[n]);
}

int ItGroup::Unlink(ItPicInGroup* pPic)
{
  for (unsigned int n=0; n<vPics.size(); n++)
    if (vPics[n] == pPic)
    {
      int nRes = pPic->pPic->DecRefCnt();
      if (nRes < 0)
      {
        // todo: internal error
      }
      vPics.erase(vPics.begin() + n);
      delete pPic;
      return nRes;
    }
  
  // todo: internal error
  return -1;
}

ItPicInGroup* ItGroup::GetPic(ItPic* pPic)
{
  for (unsigned int n=0; n<vPics.size(); n++)
    if (vPics[n]->pPic == pPic)
      return vPics[n];
  
  return 0;
}

void ItGroup::Unlink(ItGroup* pGroup)
{
  ItGroup* p = pFirstChild;

  while (p != 0 && p != pGroup)
    p = p->pNext;

  if (p == pGroup)
  {
    ItGroup* prev = pGroup->pPrev;
    ItGroup* next = pGroup->pNext;

    pGroup->pParent = 0;

    if (prev != 0)
      prev->pNext = next;
    if (next != 0)
      next->pPrev = prev;
    if (pFirstChild == pGroup)
      pFirstChild = next;

    fExpanded = (pFirstChild != 0);
  }  
}

ItGroup* ItGroup::getLastChild()
{
  ItGroup* p    = pFirstChild;
  ItGroup* pold = p;

  while (p != (ItGroup*)0)
  {
    pold = p;
    p    = p->pNext;
  }

  return(pold);
}

ItGroup* ItGroup::addNewGroup(const char* name, TreeView* pT)
{
  ItGroup* ng = new ItGData(getLastChild(), this);  
  ng->SetName(name);
  pT->AddGroup(ng);

  return(ng);
}

void ItGroup::AddGroup(ItGroup* pGroup)
{
  if (pFirstChild)
  {
    ItGroup* pLast = getLastChild();
    pLast->pNext  = pGroup;
    pGroup->pPrev = pLast;
  }
  else
  {
    pFirstChild   = pGroup;
    pGroup->pPrev = 0;
  }
  pGroup->pParent = this;
  pGroup->pNext   = 0;
}

void ItGroup::addPics(std::string            srcdir,
                      std::string            bp,
                      ItPic::eDateTimeSource eDTS, 
                      TreeView*              pT)
{
  ScanDir*        dir;
  struct stat     Attribs;              // ein Puffer für die Dateiattribute

  std::cout << GetName() << ": adding pictures from " << srcdir << "\n";

  dir = new ScanDir(srcdir.c_str());

  for (int n=0; n<dir->nNrEntries; n++)
  {
    // read Attribs
    {
      std::string tmp = srcdir;

      tmp += (dir->list)[n]->d_name;
      if (tmp[tmp.length()-1] == '/')
        tmp = tmp.substr(0, tmp.length()-1);
	  
      FileTools::stat(tmp.c_str(), &Attribs);
    }

    if (S_ISDIR(Attribs.st_mode) != 0)
    {
      // Es ist ein Verzeichnis!
      std::string tmp = (dir->list)[n]->d_name;
      if (tmp[tmp.length()-1] == '/')
        tmp = tmp.substr(0, tmp.length()-1);

      if ((tmp.compare("..") != 0) &&
          (tmp.compare(".")  != 0) &&
          (tmp.compare(".thumbnails") != 0) && // thumbnail directory used by fotoalbum
          (tmp.compare(".thumbs")     != 0)    // thumbnail directory used by mapivi
          )
      {
        std::cout << "directory: " << tmp << "\n";

        ItGroup* ng = addNewGroup(tmp.c_str(), pT);
        ng->addPics(srcdir + tmp + "/", bp, eDTS, pT);
      }
    }
    else if (S_ISREG(Attribs.st_mode) != 0)
    {
      std::string tmp = (dir->list)[n]->d_name;
      
      if ((strU(tmp).compare("THUMBS.DB") != 0))
      {
        ItPicInGroup* pic = addNewPic(pT);
        pic->pPic->setData(srcdir + (dir->list)[n]->d_name, bp, eDTS);
      }
    }
  }
  delete dir;
}

void ItGroup::addFile(std::string            srcdir,
                      std::string            bp,
                      ItPic::eDateTimeSource eDTS, 
                      TreeView*              pT)
{
  ItPicInGroup* pic = addNewPic(pT);
  pic->pPic->setData(srcdir, bp, eDTS);  
}

ItPicInGroup* ItGroup::addNewPic(TreeView* pT)
{
  ItPic* p = new ItPic();
  p->IncRefCnt();
  pT->AddPic(p);
  ItPicInGroup* pIG = new ItPicInGroup(p);
  vPics.push_back(pIG);
  return pIG; 
}

JWWDateTime ItGroup::getDT()                                              
{
  JWWDateTime min;
  JWWDateTime max;

  min.nTime = 0;
  min.nDate = 0;
  max = min;

  getMinMaxDateTime(min, max, true);

  return(min);
}

void ItGroup::SortChildren(int nMethod)
{
  bool fChanged;
  int  nC;

  // sort pictures
  if (vPics.size() > 1)
  {
    do
    {
      ItPic* p1;
      ItPic* p2;
      fChanged = false;
      for (unsigned int n=0; n<vPics.size()-1; n++)
      {
        p1 = vPics[n+0]->pPic;
        p2 = vPics[n+1]->pPic;
        
        // using some kind of function pointer would be better here...
        switch (nMethod)
        {
          case 0:
            {
              JWWDateTime dt = p2->getDT();
              nC = p1->getDT().compare( &dt );
            }
            break;
            
          default:
            nC = strcmp(p1->GetPathPtr()->c_str(), p2->GetPathPtr()->c_str());
            break;
        }
        
        if (nC > 0)
        {
          fChanged = true;
          ItPicInGroup* p = vPics[n+0];
          vPics[n+0] = vPics[n+1];
          vPics[n+1] = p;        
        }
      }
      
      for (unsigned int n=vPics.size()-1; n>0; n--)
      {
        p1 = vPics[n-0]->pPic;
        p2 = vPics[n-1]->pPic;
        
        // using some kind of function pointer would be better here...
        switch (nMethod)
        {
          case 0:
            {
              JWWDateTime dt = p2->getDT();
              nC = p1->getDT().compare( &dt );
            }
            break;
            
          default:
            nC = strcmp(p1->GetPathPtr()->c_str(), p2->GetPathPtr()->c_str());
            break;
        }
        
        if (nC < 0)
        {
          fChanged = true;
          ItPicInGroup* p = vPics[n-0];
          vPics[n-0] = vPics[n-1];
          vPics[n-1] = p;        
        }
      }    
    }
    while (fChanged);
  }
  
  // sort groups
  if (pFirstChild && nMethod == 0)
  {
    ItGroup* p1 = pFirstChild;
    ItGroup* p2;
    bool fEnd;
    do
    {
      fChanged = false;
      fEnd     = false;
      while (!fEnd)
      {
        p2 = p1->pNext;
        
        if (p2 == 0)
        {
          fEnd = true;
        }
        else
        {
          JWWDateTime dt = p2->getDT();
          nC = p1->getDT().compare( &dt );

          if (nC == 1)
          {
            fChanged = true;
            
            ItGroup* p0 = p1->pPrev;
            ItGroup* p3 = p2->pNext;
            // p0 -> p1 -> p2 -> p3
            //
            if (p0 != 0)
              p0->pNext = p2;
            p2->pPrev = p0;
            p2->pNext = p1;
            p1->pPrev = p2;
            p1->pNext = p3;
            if (p3 != 0)
              p3->pPrev = p1;
          }
          p1 = p2;
        }
      }
      
      fEnd = false;
      while (!fEnd)
      {
        p2 = p1->pPrev;
        
        if (p2 == 0)
        {
          fEnd = true;
        }
        else
        {
          JWWDateTime dt = p2->getDT();
          nC = p1->getDT().compare( &dt );
          
          if (nC == -1)
          {
            fChanged = true;
            ItGroup* p3 = p2->pPrev;
            ItGroup* p0 = p1->pNext;
            // p3 -> p2 -> p1 -> p0
            //
            if (p3 != 0)
              p3->pNext = p1;
            p1->pPrev = p3;
            p1->pNext = p2;
            p2->pPrev = p1;
            p2->pNext = p0;
            if (p0 != 0)
              p0->pPrev = p2;
          }
          p1 = p2;
        }
      }
    }
    while (fChanged);
    pFirstChild = p1;  
  }
}

void ItGroup::SortChildrenChrono()
{
  SortChildren(0);
}

void ItGroup::SortChildrenPath()
{
  SortChildren(1);
}

void ItGroup::MoveAfter(ItGroup* pDst, ItGroup* pGroup)
{
  ItGroup* pOldParent = pGroup->pParent;
  if (pOldParent)
    pOldParent->Unlink(pGroup);

  pGroup->pParent = this;
  
  ItGroup* pTail = pDst->pNext; 
  
  pDst->pNext = pGroup;
  pGroup->pPrev = pDst;
  
  pGroup->pNext = pTail;
  if (pTail)
    pTail->pPrev = pGroup;
}

void ItGroup::MoveAfter(ItPicInGroup* pDst, std::vector<ItPicInGroup*> vPicsMv)
{
  int nDst = -1;
  for (unsigned int m=0; m<vPicsMv.size(); m++)
  {
    bool done = false;
    for (unsigned int n=0; n<vPics.size() && !done; n++)
    {
      if (vPicsMv[m] == vPics[n])
        vPics.erase(vPics.begin() + n);
    }
  }
  
  for (unsigned int n=0; n<vPics.size() && nDst < 0; n++)
  {
    if (vPics[n] == pDst)
      nDst = n+1;
  }
  if (nDst < 0)
    return;
  
  for (unsigned int m=0; m<vPicsMv.size(); m++)
    vPics.insert(vPics.begin() + nDst, vPicsMv[m]);
}

std::string ItGroup::GetContentString()                                       
{
  std::string strC = "";
  for (unsigned int n=0; n<vPics.size(); n++)
  {
    strC.push_back('\n');
    strC.append(vPics[n]->pPic->strID);
  }
  
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    strC.push_back('\n');
    strC.append(p->strID);
    p = p->pNext;
  }
  
  if (strC.length())
    strC.push_back('\n');
    
  return strC;
}

void ItGroup::kw_setGroup(std::set<std::string>& kwset)
{
  // myself:
  kw_set(kwset);  
  for (unsigned int n=0; n<vPics.size(); n++)
    vPics[n]->pPic->kw_set(kwset);
  
  // children:
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    p->kw_setGroup(kwset);    
    p = p->pNext;
  }
}

void ItGroup::kw_replaceGroup(std::string from, std::string to, int mode)
{
  // myself:
  kw_replace(from, to, mode);
  for (unsigned int n=0; n<vPics.size(); n++)
    vPics[n]->pPic->kw_replace(from, to, mode);
  
  // children:
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    p->kw_replaceGroup(from, to, mode);
    p = p->pNext;
  }
}

unsigned int ItGroup::countMissingFiles(std::string basepath)
{
  unsigned int uCnt = 0;
  
  for (unsigned int n=0; n<vPics.size(); n++)
    if (FileTools::fileExists(basepath + vPics[n]->pPic->GetPath()) == false)
    {
      std::cerr << "Missing file: " << basepath << vPics[n]->pPic->GetPath() << "\n";
      uCnt++;
    }
  
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    uCnt += p->countMissingFiles(basepath);
    p = p->pNext;
  }
  
  return uCnt;
}

ItGLink* ItGroup::Search(ItGSearch* pSearch, std::set<ItPic*>* pFlatRes, std::string strBasePath, std::string strPCom, std::string strPKey)
{
  ItGLink* pML = 0;

  if (pSearch->pSP->NeedsPCom())
    strPCom += UTF8::to_lower(GetName()) + " " + UTF8::to_lower(GetComment()) + " ";
  if (pSearch->pSP->NeedsPKey())
    strPKey += UTF8::to_lower(GetKeywords());
  
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    if (p != pSearch)
    {
      ItGLink* pCL = p->Search(pSearch, pFlatRes, strBasePath, strPCom, strPKey);
      if (pCL)
      {
        if (pML == 0)
          pML = new ItGLink(0, 0, this);
        pML->AddGroup(pCL);
      }
    }
    p = p->pNext;
  }
  
  for (unsigned int n=0; n<vPics.size(); n++)
  {
    if (pSearch->pSP->Run(vPics[n], this, strBasePath, strPCom, strPKey))
    {
      if (pFlatRes)
        pFlatRes->insert(vPics[n]->pPic);
      else
      {
        if (pML == 0)          
          pML = new ItGLink(0, 0, this);
        pML->vPics.push_back(new ItPicInGroup(vPics[n]->pPic));
      }
    }
  }
    
  return pML;
}

void ItGroup::DebugShowStructure(int level)
{
  for (int n=0; n<level; n++)
    std::cout << " ";
  std::cout << /*this << " " <<*/ GetName() /* << ". par=" << pParent << ", prev=" << pPrev */ << "\n";

  for (unsigned int n=0; n<vPics.size(); n++)
  {
    for (int m=0; m<level+2; m++)
      std::cout << " ";
    std::cout << vPics[n]->pPic->GetPath() << "\n";
  }  
  
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    p->DebugShowStructure(level+2);
    p = p->pNext;
  }
}

void ItGroup::DeleteMyContent()
{
  unsigned int size = vPics.size();
  for (unsigned int n=0; n<size; n++)
    delete vPics[n];
  vPics.clear();
  
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    p->DeleteMyContent();

    ItGroup* next = p->pNext;
    delete p;
    p = next;
  }  
  pFirstChild = 0;
}

ItGroup::~ItGroup() 
{
  // std::cout << "ItGroup::~ItGroup()\n";
}

std::string ItGroup::FindFile(std::string basepath,
                              std::string filename)                       
{
  for (unsigned int n=0; n<vPics.size(); n++)
  {
    std::string path = basepath + vPics[n]->pPic->GetPath();
    path = path.substr(0, path.rfind('/')+1) + filename;
    if (FileTools::fileExists(path))
      return(path);
  }
  
  ItGroup* p = pFirstChild;
  while (p != 0)
  {
    std::string erg = p->FindFile(basepath, filename);
    if (erg.length())
      return(erg);    
    p = p->pNext;
  }

  return("");
}



ItGRoot::ItGRoot(std::string strID, std::string strName) : ItGroup(0, 0)
{
  this->strID   = strID;
  this->strName = strName;
}


ItGData::ItGData(SimpleXMLTransfer* xIn, int nVer)
{
  strID       = xIn->attribute("UID");
  strKeywords = xIn->attribute("k", "");  kw_fix(strKeywords);
  strComment  = xIn->attribute("c", "");
  strName     = xIn->attribute("t");

  fExpanded = false;
}

ItGData::ItGData(SimpleXMLTransfer* xIn) : ItGroup(0, 0)
{
  strID       = xIn->getName();
  strKeywords = xIn->attribute("k", "");  kw_fix(strKeywords);
  strComment  = xIn->attribute("c", "");
  strName     = xIn->attribute("t");

  fExpanded = false;
}

ItGData::ItGData(ItGroup* pPrev, ItGroup* pParent) : ItGroup(pPrev, pParent)
{
  strKeywords = "";
  strComment  = "";
  strName     = "";
}

void ItGData::SaveToXML(SimpleXMLTransfer* xGroups) 
{
  SimpleXMLTransfer* xMe = new SimpleXMLTransfer();
  xMe->setName(strID);
  xGroups->addChild(xMe);

  xMe->addAttribute(XMLNAME_SAVE_TITLE, strName);
 
  std::string strKeywordsFile = GetKeywordsFile();  
  if (!strKeywordsFile.empty())
    xMe->addAttribute(XMLNAME_SAVE_KEYWORDS, strKeywordsFile);
  
  if (!strComment.empty())
    xMe->addAttribute(XMLNAME_SAVE_COMMENT, strComment);

  std::string strC = GetContentString();
  if (strC.length())
    xMe->setContent(strC);
}

ItGData::~ItGData()
{
  // std::cout << "ItGData::~ItGData(): " << GetName() << "\n";
}


ItGSearch::ItGSearch(SimpleXMLTransfer* xIn) : ItGData(xIn)
{
  strExpr = xIn->attribute("s");
  nFlat   = xIn->attributeAsInt("f");  
  pSrc = 0;  
  pSP  = new SearchProg(strExpr);
}

ItGSearch::ItGSearch(ItGroup* pPrev, ItGroup* pParent, ItGroup* pSrc, std::string strExpr, int nFlat) : ItGData(pPrev, pParent)
{
  strKeywords = "";
  strComment  = "";
  strName     = "new search";
  this->pSrc  = pSrc;  
  this->strExpr = strExpr;
  this->nFlat   = nFlat;
  pSP = 0;
}

void ItGSearch::SaveToXML(SimpleXMLTransfer* xGroups)
{
  SimpleXMLTransfer* xMe = new SimpleXMLTransfer();
  xMe->setName(strID);
  xGroups->addChild(xMe);

  xMe->addAttribute(XMLNAME_SAVE_TITLE, strName);
  
  std::string strKeywordsFile = GetKeywordsFile();  
  if (!strKeywordsFile.empty())
    xMe->addAttribute(XMLNAME_SAVE_KEYWORDS, strKeywordsFile);
  
  if (!strComment.empty())
    xMe->addAttribute(XMLNAME_SAVE_COMMENT, strComment);
  
  xMe->addAttribute(XMLNAME_SAVE_SEARCH, strExpr);
  xMe->addAttribute(XMLNAME_SAVE_FLAGS,  nFlat);
  if (pSrc)
    xMe->addAttribute(XMLNAME_SAVE_LINK,   pSrc->GetID());
  else
    xMe->addAttribute(XMLNAME_SAVE_LINK,   "0");
}

void ItGSearch::Refresh(std::string strBasePath)
{
  DeleteMyContent();
  
  if (pSrc && pSP && pSP->GetErr().length() == 0)
  {
    HiResTime hrt;
    hrt.Capture();
    
    std::string strPCom = "";
    if (pSP->NeedsPCom())
    {
      ItGroup* p = pSrc->pParent;
      while (p)
      {
        strPCom += UTF8::to_lower(p->GetName()) + " " + UTF8::to_lower(p->GetComment()) + " ";
        p = p->pParent;
      }
    }
    std::string strPKey = "";
    if (pSP->NeedsPKey())
    {
      ItGroup* p = pSrc->pParent;
      while (p)
      {
        strPKey += UTF8::to_lower(p->GetKeywords());
        p = p->pParent;
      }
    }
    
    if (nFlat)
    {
      std::set<ItPic*> sFlatRes;
      // flat result, no duplicates
      pSrc->Search(this, &sFlatRes, strBasePath, strPCom, strPKey);
      for (std::set<ItPic*>::iterator it=sFlatRes.begin(); it != sFlatRes.end(); it++)
        vPics.push_back(new ItPicInGroup(*it));
    }
    else
    {
      // hierarchical result
      ItGLink* pCL = pSrc->Search(this, 0, strBasePath, strPCom, strPKey);
      if (pCL)
      {
        if (true)
        {
          // Alle Inhalte der Gruppe pCL muss ich übernehmen und pCL wieder löschen
          while (pCL->pFirstChild)
          {
            ItGroup* p = pCL->pFirstChild;
            pCL->Unlink(p);
            AddGroup(p);
          }
          while (pCL->vPics.size())
          {
            ItPic* pP = pCL->vPics[0]->pPic;
            pCL->Unlink(pCL->vPics[0]);
            vPics.push_back(new ItPicInGroup(pP));
            pP->IncRefCnt();
          }            
          delete pCL;
        }
        else
        {
          // Auf diese Art und Weise wäre die durchsuchte Gruppe selbst im Ergebnis enthalten, 
          // was aber nicht gewünscht ist:
          AddGroup(pCL);
        }
      }
    }

    std::cout << "Search completed in " << hrt.GetMSecSinceCapture() << " ms.\n";
  }
}

void ItGSearch::SetSearchProg(SearchProg* pSP)
{
  if (this->pSP) 
    delete this->pSP; 
  this->pSP = pSP; 
}

int ItGSearch::SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups)
{
  std::string strLink = xIn->attribute("l");
  
  std::map<std::string, ItGroup*>::iterator itG = mapGroups.find(strLink);
  if (itG == mapGroups.end())
  {
    if (strLink.compare("0") == 0)
      return 0;
    std::cerr << __LINE__ << ", " << GetName() << ", " << GetID() << ", Link target not found: *" << strLink << "*\n";
    return 1;
  }
  else
  {
    pSrc = itG->second;
    return 0;
  }
}

void ItGSearch::Deleted(ItGroup* p)
{ 
  if (p == pSrc)
    pSrc = 0;
  
  DeleteMyContent();
}

ItGSearch::~ItGSearch()
{
  // std::cout << "ItGSearch::~ItGSearch(): " << GetName() << "\n";
}






ItGLink::ItGLink(ItGroup* pPrev, ItGroup* pParent, ItGroup* pLink) : ItGroup(pPrev, pParent)
{
  this->pLink = pLink;
  fExpanded = false;
}

ItGLink::~ItGLink() 
{
  // std::cout << "ItGLink::~ItGLink(): " << GetName() << "\n";
}













ItGLinkRef::ItGLinkRef(SimpleXMLTransfer* xIn, int nVer) : ItGLink(0, 0, 0)
{
  strID = xIn->attribute("UID");
  fExpanded = false;
  pLink = 0;
}


ItGLinkRef::ItGLinkRef(SimpleXMLTransfer* xIn) : ItGLink(0, 0, 0)
{
  strID = xIn->getName();
  fExpanded = false;
  pLink = 0;
}

int ItGLinkRef::SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups)
{
  std::string strLink = xIn->attribute("l");
  
  std::map<std::string, ItGroup*>::iterator itG = mapGroups.find(strLink);
  if (itG == mapGroups.end())
  {
    std::cerr << __LINE__ << ", " << GetName() << ", " << GetID() << ", Link target not found: *" << strLink << "*\n";
    return 1;
  }
  else
  {
    pLink = itG->second;
    itG->second->IncRefCnt();
    return 0;
  }
}

ItGLinkRef::ItGLinkRef(ItGroup* pPrev, ItGroup* pParent, ItGroup* pLink) : ItGLink(pPrev, pParent, pLink)
{
  pLink->IncRefCnt();
}

ItGLinkRef::~ItGLinkRef()
{
  std::cout << "ItGLinkRef::~ItGLinkRef(): " << GetName() << "\n";
  if (pLink)
    pLink->DecRefCnt();
}

void ItGLinkRef::SaveToXML(SimpleXMLTransfer* xGroups) 
{
  SimpleXMLTransfer* xMe = new SimpleXMLTransfer();
  xMe->setName(strID);
  xGroups->addChild(xMe);

  xMe->addAttribute(XMLNAME_SAVE_LINK, pLink->strID);
  
  std::string strC = GetContentString();
  if (strC.length())
    xMe->setContent(strC);
}


