/*
 * Copyright (C) 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#ifndef BASE_GROUP_H
# define BASE_GROUP_H

# include <string>
# include <map>
# include <vector>
# include "../shared/SimpleXMLTransfer.h"
# include "pic.h"
# include "meta.h"

class TreeView;
class SearchProg;
class ItGLink;
class ItGSearch;


class ItGroup : public ItMeta
{
public:
  int nRefCnt;
  int IncRefCnt() { return(++nRefCnt); };
  int DecRefCnt() { return(--nRefCnt); };
  void GetMarkedPics(std::vector<ItPicInGroup*>& vMarkedPics);
  
  /**
   * add a new group at the end of my children
   */
  ItGroup* addNewGroup(const char* name, TreeView* pT);
    
  void AddGroup(ItGroup* pGroup);

  /**
   * returns pointer to last child
   */
  ItGroup* getLastChild();
  
  /**
   * Returns pPic's reference count after unlinking
   */
  int Unlink(ItPicInGroup* pPic);
  void Unlink(ItGroup* pGroup);

  void addPics(std::string              srcdir,
               std::string              bp,
               ItPic::eDateTimeSource   eDTS,
               TreeView*                pT);
  void addFile(std::string              srcdir,
               std::string              bp,
               ItPic::eDateTimeSource   eDTS,
               TreeView*                pT);
  
  /**
   * add a new picture
   */
  ItPicInGroup* addNewPic(TreeView* pT);
  
  void SortChildrenChrono();
  void SortChildrenPath();
  
  void MoveAfter(ItGroup* pDst, ItGroup* pGroup);
  void MoveAfter(ItPicInGroup* pDst, std::vector<ItPicInGroup*> vPics);
  
  std::string GetContentString();
  virtual void SaveToXML(SimpleXMLTransfer* xGroups) {};
  
  /**
   * Adds keywords to the list recursively (no dupes, alphabetically)
   */
  void kw_setGroup(std::set<std::string>& kwset);

  /**
   * replace keywords recursively
   */
  void kw_replaceGroup(std::string from, std::string to, int mode);
  
  /**
   * returns the number of missing files -- 0 if everything is fine
   */
  unsigned int countMissingFiles(std::string basepath);

  ItGLink* Search(ItGSearch* pSearch, std::set<ItPic*>* pFlatRes, std::string strBasePath, std::string strPCom, std::string strPKey);
  
  void DebugShowStructure(int level);

  virtual int SearchType() { return 0; };

  std::string FindFile(std::string basepath,
                       std::string filename);

public:
  static ItGroup* Factory(SimpleXMLTransfer* xIn);
  ItGroup();
  ItGroup(ItGroup* pPrev, ItGroup* pParent);
  virtual ~ItGroup();
  
  /**
   * min and max have to be reset before calling this method!
   * It returns the number of non-groups.
   */
  unsigned int getMinMaxDateTime(JWWDateTime& min,
                                 JWWDateTime& max,
                                 bool         fRecursive);
  JWWDateTime getDT();

  void DeleteMyContent();
  
public:
  virtual std::string GetName() = 0;
  virtual void SetName(std::string strName) {};
  virtual const std::string* GetNamePtr()  { return 0; };
  
public:
  int SetChildren(SimpleXMLTransfer* xIn, std::map<std::string, ItPic*>& mapPics, std::map<std::string, ItGroup*>& mapGroups);

  std::vector<ItPicInGroup*> vPics;
  ItPicInGroup* GetPic(ItPic* pPic);  
  
  ItPicInGroup* GetPicNext(ItPic* pPic);
  ItPicInGroup* GetPicPrev(ItPic* pPic);
  
  /**
   * Pointer to the group to which I belong.
   * this is public to allow for a fast @ref treeDepth() (without 
   * recursive method calls).
   */
  ItGroup* pParent;
  
  ItGroup* pFirstChild;
  
  /**
   * Previous item in the same group.
   */
  ItGroup* pPrev;
  
  /**
   * Next item in the same group.
   * This is public to allow for fast group operations.
   */
  ItGroup* pNext;
  void SetParent(ItGroup* pGroup) { pParent = pGroup; };
  
  /**
   * Currently expanded in tree view?
   */
  bool fExpanded;
  
  /**
   * Jedes Element muss wissen, wie weit es eingerückt werden
   * muss. Dazu muss man die Anzahl der Schritte über @ref pParent bis zum
   * Null-Element zählen. Es werden nicht viele gleichzeitig dargestellt, daher
   * kann man das jedesmal zählen. Alternativ könnte sich auch jede @ref ItGroup
   * diese Entfernung merken.
   */
  int treeDepth();
  
  /**
   * 
   */
  bool canExpand() { return (pFirstChild != 0); };
  
protected:
  virtual int SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups) = 0;
  
  void SortChildren(int nMethod);
};

class ItGRoot : public ItGroup
{
public:
  ItGRoot(std::string strID, std::string strName);
  virtual ~ItGRoot() {};
  
  virtual std::string GetKeywords() { return ""; };
  virtual std::string GetComment()  { return "";  };
  virtual const std::string* GetCommentPtr()   { return 0;  };
  virtual const std::string* GetKeywordsPtr()  { return 0; };
  virtual std::string GetName()     { return strName; };
  virtual const std::string* GetNamePtr()  { return &strName; };
  
  virtual int SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups) { return 0; };
private:
  std::string strName;
};

class ItGData : public ItGroup
{
public:
  std::string strName;

protected:
  std::string strKeywords;
  std::string strComment;
public:
  virtual std::string GetComment()  { return strComment;  };
  virtual std::string GetKeywords() { return strKeywords; };
  virtual const std::string* GetCommentPtr()   { return &strComment;  };
  virtual const std::string* GetKeywordsPtr()  { return &strKeywords; };
  virtual void SetComment(std::string strNew)  { strComment  = strNew; };
  virtual void SetKeywords(std::string strNew) { strKeywords = strNew; kw_fix(strKeywords); };  
  
public:
  ItGData(SimpleXMLTransfer* xIn, int nVer);
  ItGData(SimpleXMLTransfer* xIn);
  ItGData(ItGroup* pPrev, ItGroup* pParent);
  virtual ~ItGData();
  
  virtual std::string GetName()     { return strName;     };
  virtual void SetName(std::string strName) { this->strName = strName; };
  virtual const std::string* GetNamePtr()  { return &strName; };

  virtual int SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups) { return 0; };
  virtual void SaveToXML(SimpleXMLTransfer* xGroups);
};

class ItGLink : public ItGroup
{
public:
  ItGroup* pLink;
  
public:
  ItGLink(ItGroup* pPrev, ItGroup* pParent, ItGroup* pLink);
  virtual ~ItGLink();
  virtual int SearchType() { return 2; };

  virtual std::string GetKeywords() { return pLink->GetKeywords(); };
  virtual std::string GetComment()  { return pLink->GetComment();  };
  virtual const std::string* GetCommentPtr()   { return pLink->GetCommentPtr();  };
  virtual const std::string* GetKeywordsPtr()  { return pLink->GetKeywordsPtr(); };
  virtual void SetComment(std::string strNew)  { pLink->SetComment(strNew);  };
  virtual void SetKeywords(std::string strNew) { pLink->SetKeywords(strNew); };
  
  virtual std::string GetName()             { return pLink->GetName(); };
  virtual void SetName(std::string strName) { pLink->SetName(strName); };
  virtual const std::string* GetNamePtr()   { return pLink->GetNamePtr(); };
  
  virtual int SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups) { return 0; };
};

class ItGLinkRef : public ItGLink
{
public:
  ItGLinkRef(SimpleXMLTransfer* xIn, int nVer);
  ItGLinkRef(SimpleXMLTransfer* xIn);
  ItGLinkRef(ItGroup* pPrev, ItGroup* pParent, ItGroup* pLink);
  virtual ~ItGLinkRef();
  virtual int SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups);
  virtual void SaveToXML(SimpleXMLTransfer* xGroups);
  virtual int SearchType() { return 0; };
};

class ItGSearch : public ItGData
{
public:
  ItGSearch(SimpleXMLTransfer* xIn);
  ItGSearch(ItGroup* pPrev, ItGroup* pParent, ItGroup* pSrc, std::string strExpr, int nFlat);
  virtual ~ItGSearch();
  virtual int SearchType() { return 1; };
  
  virtual void SaveToXML(SimpleXMLTransfer* xGroups);
  virtual int SetLink(SimpleXMLTransfer* xIn, std::map<std::string, ItGroup*>& mapGroups);

  void SetSearchProg(SearchProg* pSP);
  void Refresh(std::string strBasePath);
  
  void Deleted(ItPic* p)   { DeleteMyContent(); };
  void Deleted(ItGroup* p);
  
public:
  ItGroup* pSrc;
  std::string strExpr;
  int nFlat;
  SearchProg* pSP;
};

#endif
