/*
 * Copyright (C) 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "pic.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include "xmldefs.h"
#include "basetools.h"
#include "keyword.h"
#include "../shared/lib_conversions.h"
#include "../shared/filetools.h"
#include "exif.h"

ItPic::ItPic(SimpleXMLTransfer* xIn, std::string strPathPrefix)
{
  strID       = xIn->getName();
  strKeywords = xIn->attribute("k", "");  kw_fix(strKeywords);
  strComment  = xIn->attribute("c", "");
  dt.nTime    = xIn->attributeAsInt(XMLNAME_TIME, 0);
  dt.nDate    = xIn->attributeAsInt(XMLNAME_DATE, 0);
  strPath     = strPathPrefix + xIn->attribute("p");
  DecodePath(strPath);

  nRefCnt = 0;
}

ItPic::ItPic(SimpleXMLTransfer* xIn)
{
  strID       = xIn->attribute("UID");
  strKeywords = xIn->attribute("k", "");  kw_fix(strKeywords);
  strComment  = xIn->attribute("c", "");
  dt.nTime    = xIn->attributeAsInt(XMLNAME_TIME, 0);
  dt.nDate    = xIn->attributeAsInt(XMLNAME_DATE, 0);
  strPath     = xIn->attribute("p");
  DecodePath(strPath);

  nRefCnt = 0;
}

ItPic::ItPic()
{
  strID    = "";
  nRefCnt  = 0;
  dt.nTime = 0;
  dt.nDate = 0;
}

void ItPic::DecodePath(std::string& path)
{
  int length = path.length();
  int o = 0;
  char ch;
  for (int i=0; i<length; )
  {
    ch = path.at(i);
    if (ch != '#')
    {
      path[o++] = ch;
      i++;
    }
    else
    {
      path[o++] = (char)( (hex2int(path.at(i+1))<<4) + hex2int(path.at(i+2)) );
      i+= 3;
    }       
  }
  if (o < length)
    path.resize(o);
}

std::string ItPic::EncodePath(std::string path)
{
  int length = path.length();
  char ch;
  std::string strBuf;
  
  for (int i = 0; i < length; i++)
  {
    ch = path.at(i);
    
    if (ch != ' ' && (ch < '(' || ch > '~' || ch == '<' || ch == '>'))
    {
      if (strBuf.empty())
        strBuf = path.substr(0, i);
      strBuf.push_back('#');
      strBuf.append(itoHexStr( (int)((unsigned char)(ch)) ));
    }
    else if (!strBuf.empty())
      strBuf.push_back(ch);    
  }
  
  if (strBuf.empty())
    return(path);
  else
    return(strBuf);
}

std::string ItPic::getDescr(int level, bool fShort, const std::string strBasepath)
{
  const unsigned int maxlen = 50;
  
  std::string tmp = "";
  bool        fDateTime = false;

  if (vRating.size() && vRating[0] != eRatingNone && !fShort)
  {
    tmp += "<p>";
    if (vRating[0] == eRatingWorst)
      tmp += "X</p>";
    else
      tmp += itoStr(vRating[0], ' ', 1) + "</p>";
  }
  
  if (level >= 1 && (hasDate() || hasTime())) 
  {
    if (!fShort)
      tmp += "<p>";
  
    if (hasDate())
    {
      tmp += getDateString() + " ";
      fDateTime = true;
    }
    
    if (hasTime())
    {
      tmp += getTimeString();
      fDateTime = true;
    }

    if (!fShort)
      tmp += "</p>";
  }
  
  if (level >= 2)
  {
    if (fDateTime && strComment.length() > 0)
    {
      tmp += "\n";
    }
    
    if (strComment.length())
    {
      if (fShort)
      {
        if (strComment.length() > maxlen)
          tmp += strComment.substr(0, maxlen-3) + "...";
        else
          tmp += strComment;
      }
      else
        tmp += BaseTools::GetAsHTML(strComment);
    }
  }

  if (level >= 3)
  {
    std::vector<std::string> kwlist;
    unsigned int size;
    Keywords::list(kwlist, strKeywords);
    size = kwlist.size();
    if (size)
    {
      if (!fShort)
        tmp += "<p>";
      
      if (fShort && size > 10)
        size = 10;
      if (tmp.length() && kwlist.size() > 0)
        tmp += "\n";
      for (unsigned int n=0; n<size; n++)
        tmp += "\n" + kwlist[n];
      if (size != kwlist.size())
        tmp += "  ...";
      
      if (!fShort)
        tmp += "</p>";
    }
  }
  
  if (level >= 4)
  {
    std::string t = GetPath();
    if (t.length())
    {
      tmp += "\n";
      if (!fShort)
        tmp += "<p>";      
      tmp += t;
      if (!fShort)
        tmp += "</p>";      
    }
  }
  
  if (level >= 5)
  {
    std::string t = "";
    {
      std::string mypath = strBasepath + strPath;
      ExifInfo*   exif   = new ExifInfo(mypath);
      if (exif->hasEXIF())
      {
        t = "----------- EXIF -------------\n";
        t += exif->getInfo();
      }
      delete exif;      
    }

    // replace newline with <br/>
    std::string tout = "";
    bool fNL = false;
    for (std::string::size_type n=0; n<t.length(); n++)
    {
      if (t[n] == '\n')
      {
        if (!fNL)
          tout += "<br/>";
        fNL = true;
      }
      else
      {
        fNL = false;
        tout += t[n];
      }
    }
    if (t.length())
    {
      tmp += "\n";
      if (fShort)
        tmp += t;
      else
        tmp += "<p>" + tout + "</p>";      
    }
  }
  
  return(tmp);
}

void ItPic::setData(std::string       newpath,
                    std::string       basepath,
                    eDateTimeSource   eDTS)
{
  bool fSet = false;
  ExifInfo* exif = new ExifInfo(newpath);
  
  // set time and date
  if (eDTS != eFILE)
  {
    // try exif
    if (exif->hasEXIF())
    {
      std::string tmp = exif->getDateTime();
      if (tmp.length() == 19)
      {
        dt.setDate(tmp.substr(0, 10).c_str());
        dt.setTime(tmp.substr(11).c_str());
        fSet = true;
      }      
    }
  }
  
  if (!fSet && eDTS != eEXIF)
  {
    // take remaining data from file
    struct stat     Attribs;
    struct tm       timestruct;
    
    FileTools::stat(newpath.c_str(), &Attribs);

#ifndef WIN32
    localtime_r(&Attribs.st_mtime, &timestruct);
#else
    // todo: is localtime without _r ok?
#endif
            
    dt.nDate = JWWDateTime::assembleDate(1900 + timestruct.tm_year, timestruct.tm_mon+1, timestruct.tm_mday);
    
    if (eDTS != eEXIFFILEDATE)
      dt.nTime = JWWDateTime::assembleTime(timestruct.tm_hour, timestruct.tm_min, timestruct.tm_sec);
  }
  
  // read additional data (comment, keywords, ...)
  strComment  = exif->ReadComment("");
  strKeywords = "";    
  {
    std::set<std::string> kwset;
    exif->ReadKeywords(kwset);
    kw_add(kwset);
  }
    
  delete exif;        
  strPath = newpath.substr(basepath.length());
}

void ItPic::SaveToXML(std::string& strPathPrefix, SimpleXMLTransfer* xPics)
{
  // clean up strKeywords
  std::string strKeywordsFile = GetKeywordsFile();
  
  //  
  std::string strPP = FileTools::path(strPath);
  if (strPP.compare(strPathPrefix) != 0)
  {
    strPathPrefix = strPP;
    SimpleXMLTransfer* xI = new SimpleXMLTransfer();
    xI->setName("P");
    xI->addAttribute("p", EncodePath(strPathPrefix));
    xPics->addChild(xI);
  }
  
  SimpleXMLTransfer* xI = new SimpleXMLTransfer();
  xI->setName(strID);
  xI->addAttribute("p", EncodePath(FileTools::name(strPath)));
  
  if (!strKeywordsFile.empty())
    xI->addAttribute(XMLNAME_SAVE_KEYWORDS, strKeywordsFile);
  
  if (!strComment.empty())
    xI->addAttribute(XMLNAME_SAVE_COMMENT,  strComment);
  
  if (dt.nTime != 0)
    xI->addAttribute(XMLNAME_SAVE_TIME,     dt.nTime);
  if (dt.nDate != 0)
    xI->addAttribute(XMLNAME_SAVE_DATE,     dt.nDate);
  
  xPics->addChild(xI);
}

void ItPic::ChangeBasepath(std::string oldbp, std::string newbp)
{
  std::string tmp = oldbp + strPath;
  strPath = tmp.substr(newbp.length());
}

void ItPic::SetRating(int nRate)
{
  if (vRating.size() == 0)
    vRating.resize(1, nRate);
  else
    vRating[0] = nRate;
}

int ItPic::FileRename(std::string strBase,
                      std::string strNew)
{
  FileTools::move(strBase+strNew, strBase+strPath);
  if (FileTools::fileExists(strBase+strNew))
  {
    strPath = strNew;
    return 0;
  }
  else
    return -1;
}





ItPicInGroup::ItPicInGroup(ItPic* pPic)
{
  this->pPic = pPic;
  fMarked    = false;
}

ItPicInGroup::ItPicInGroup()
{
  this->pPic = 0;
  fMarked    = false;
}

