/*
 * Copyright (C) 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "searchprog.h"

#include <set>
#include <iostream>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>

#include "pic.h"
#include "group.h"
#include "basetools.h"
#include "strsearchgroup.h"
#include "../shared/lib_conversions.h"
#include "../shared/filetools.h"
#include "../shared/utf8.h"



std::map<std::string, SearchFun*> SearchProg::mapFunsNoArg;
std::map<std::string, SearchFun*> SearchProg::mapFunsStrArg;

#include "search_testfun.cpp"
#include "search_sf.cpp"


void SearchContext::ReadExif(std::string strPath)
{
  if (!pExif)
    pExif = new ExifInfo(strPath);
}


TestStack::eDataType TestStack::GetType(int nOff)
{
  int size = stackType.size();
  if (nOff+1 > size)
    return eErr;
  else
    return stackType[size-1-nOff];
}

void TestStack::Push(TestStack::eDataType type, std::string strExplain)
{
  stackType.push_back(type);
  stackExplain.push_back(strExplain);
}

TestStack::eDataType TestStack::Pop(std::string& strExplain)
{
  int size = stackExplain.size();
  if (size > 0)
    strExplain = stackExplain[size-1];
  else
    strExplain = "error!";  
  
  TestStack::eDataType tmp = GetType(0);
  if (tmp != eErr)
  {
    stackType.pop_back();
    stackExplain.pop_back();
  }
  return tmp;
}

void TestStack::Print(std::string& strOut)
{
  strOut.append("<table>");
  for (int n=stackType.size()-1; n>=0; n--)
  {
    strOut.append("<tr><td>");
    if (stackType.size()-n-1 == 0) // als workaround für schlechtes itoStr
      strOut.append("[  0]</td><td>");
    else
    {
      strOut.append("[-");
      strOut.append(itoStr((int)(stackType.size())-n-1, ' ', 2));
      strOut.append("]");
      strOut.append("</td><td>");
    }
    switch (stackType[n])
    {
      case eInt:  strOut.append("int");    break;
      case eTime: strOut.append("time");   break;
      case eDate: strOut.append("date");   break;
      case eStr:  strOut.append("string"); break;
      default:    strOut.append("error!"); break;
    }
    strOut.append("</td><td>");
    strOut.append(stackExplain[n]);
    strOut.append("</td></tr>\n");
  }
  strOut.append("</table>");
}

SearchProg::SearchProg(std::string strExpr)
{
  if (mapFunsNoArg.size() == 0)
  {
    /**
     * String operations (like com.l) may be expensive, so they
     * shouldn't be repeated.
     * 
     * With str.find taking the pointer from the stack:
     *   name.l dup str.find("kastanie") swap dup str.find("pfalz") swap str.find("samm") * *
     * With str.find leaving the pointer on the stack:
     *   name.l str.find("kastanie") swap str.find("pfalz") swap str.find("samm") swap pop * *
     * The second version seems easier to understand, but both are equally fast. A
     * simple search can be written as
     *   name.l str.find("kastanie")
     * in both cases. However, if str.find doesn't take the pointer, this will leave
     * a stack with two items (forcing the user to write programs which result in a 
     * stack size of one might prevent some errors).
     * 
     * Another way of doing it: str.find swaps pointer and result by itself. In this
     * case, the simple and the compound expression become:
     *   name.l str.find("kastanie") str.find("pfalz") str.find("samm") pop * *
     *   name.l str.find("kastanie") pop
     * This is an advantage in both execution speed and readability in case of the
     * compound expression and only a minor disadvantage in case of the simple 
     * expression.
     */
    
    /**
     * More ideas for commands:
     *   str.cat:   concatenate two strings from stack, leaves result as temporary string
     *   get picture width
     *   get picture height
     *   get number of pixels in picture
     *   file.suf: get file suffix
     *   get compression level (jpeg) or bitrate (movie)
     */
    
    mapFunsNoArg.insert(std::make_pair("key",         new SearchFun(tf_push_str, sf_key)));
    mapFunsNoArg.insert(std::make_pair("key.l",       new SearchFun(tf_push_str, sf_key_l, 1)));
    mapFunsNoArg.insert(std::make_pair("key.num",     new SearchFun(tf_push_int, sf_key_num)));
    mapFunsNoArg.insert(std::make_pair("key.rmdup",   new SearchFun(tf_nop,      sf_key_rmdup)));

    mapFunsNoArg.insert(std::make_pair("com",         new SearchFun(tf_push_str, sf_com)));
    mapFunsNoArg.insert(std::make_pair("com.l",       new SearchFun(tf_push_str, sf_com_l, 1)));
    
    mapFunsNoArg.insert(std::make_pair("pcom.l",      new SearchFun(tf_push_str, sf_pcom_l, 1)));
    mapFunsNoArg.insert(std::make_pair("pkey.l",      new SearchFun(tf_push_str, sf_pkey_l, 1)));
    
    mapFunsNoArg.insert(std::make_pair("name",        new SearchFun(tf_push_str, sf_name)));
    mapFunsNoArg.insert(std::make_pair("name.l",      new SearchFun(tf_push_str, sf_name_l, 1)));
    
    mapFunsNoArg.insert(std::make_pair("ref.num",     new SearchFun(tf_push_int, sf_ref_num)));
    
    mapFunsNoArg.insert(std::make_pair("date",        new SearchFun(tf_push_date, sf_date)));
    mapFunsNoArg.insert(std::make_pair("time",        new SearchFun(tf_push_time, sf_time)));
    
    mapFunsStrArg.insert(std::make_pair("childof",     new SearchFun(tf_push_int, sf_childof)));
    mapFunsStrArg.insert(std::make_pair("childof.r",   new SearchFun(tf_push_int, sf_childof_r)));
    
    mapFunsNoArg.insert(std::make_pair("rate.best",   new SearchFun(tf_push_int, sf_rate_best)));
    mapFunsNoArg.insert(std::make_pair("rate.worst",  new SearchFun(tf_push_int, sf_rate_worst)));
    mapFunsNoArg.insert(std::make_pair("rate.mean10", new SearchFun(tf_push_int, sf_rate_mean10)));
    mapFunsNoArg.insert(std::make_pair("rate.num",    new SearchFun(tf_push_int, sf_rate_num)));
    mapFunsNoArg.insert(std::make_pair("rate.get",    new SearchFun(tf_push_int, sf_rate_get)));
    mapFunsNoArg.insert(std::make_pair("rate.set",    new SearchFun(tf_pop_int,  sf_rate_set)));

    mapFunsNoArg.insert(std::make_pair("mark.set",    new SearchFun(tf_pop_int, sf_mark_set)));
    mapFunsNoArg.insert(std::make_pair("mark.reset",  new SearchFun(tf_pop_int, sf_mark_reset)));
    mapFunsNoArg.insert(std::make_pair("mark.copy",   new SearchFun(tf_pop_int, sf_mark_copy)));
        
    mapFunsNoArg.insert(std::make_pair("id",          new SearchFun(tf_push_str, sf_id)));
    
    mapFunsNoArg.insert(std::make_pair("path",        new SearchFun(tf_push_str, sf_path)));

    mapFunsStrArg.insert(std::make_pair("str.find",   new SearchFun(tf_str_eval, sf_str_find)));
    mapFunsStrArg.insert(std::make_pair("str.count",  new SearchFun(tf_str_eval, sf_str_count)));
    mapFunsNoArg.insert(std::make_pair("str.len",     new SearchFun(tf_str_eval, sf_str_len)));
    mapFunsNoArg.insert(std::make_pair("str.cat",     new SearchFun(tf_str_cat,  sf_str_cat, 1)));
    
    mapFunsNoArg.insert(std::make_pair("file.size",   new SearchFun(tf_push_int,  sf_file_size))); // liefert -1, wenn die Datei nicht vorhanden ist!
    mapFunsNoArg.insert(std::make_pair("file.type",   new SearchFun(tf_push_int,  sf_file_type)));
    mapFunsNoArg.insert(std::make_pair("file.date",   new SearchFun(tf_push_date, sf_file_date)));

    mapFunsNoArg.insert(std::make_pair("dup",         new SearchFun(tf_dup,  sf_dup)));
    mapFunsNoArg.insert(std::make_pair("swap",        new SearchFun(tf_swap, sf_swap)));
    mapFunsNoArg.insert(std::make_pair("pop",         new SearchFun(tf_pop,  sf_pop)));
    
    // true and false can be used to skip computationally expensive parts of an expression, 
    // for example:
    //   date X < date Y * ! false ! key.l str.find("bla") *
    // In case the 'easy' part of the expression means that the expression can never
    // be true, false/true may be used to take a shorter route.
    mapFunsNoArg.insert(std::make_pair("true",        new SearchFun(tf_test_int, sf_true)));
    mapFunsNoArg.insert(std::make_pair("false",       new SearchFun(tf_test_int, sf_false)));
    
    mapFunsNoArg.insert(std::make_pair("!",           new SearchFun(tf_unary,    sf_NOT)));
    mapFunsNoArg.insert(std::make_pair(">",           new SearchFun(tf_cmp,      sf_GREATER)));
    mapFunsNoArg.insert(std::make_pair("<",           new SearchFun(tf_cmp,      sf_SMALLER)));
    mapFunsNoArg.insert(std::make_pair("==",          new SearchFun(tf_cmp,      sf_EQUAL)));
    mapFunsNoArg.insert(std::make_pair("+",           new SearchFun(tf_binary,   sf_ADD)));
    mapFunsNoArg.insert(std::make_pair("-",           new SearchFun(tf_binary,   sf_SUB)));
    mapFunsNoArg.insert(std::make_pair("*",           new SearchFun(tf_binary,   sf_MUL)));    
    mapFunsNoArg.insert(std::make_pair("min",         new SearchFun(tf_binary,   sf_MIN)));
    mapFunsNoArg.insert(std::make_pair("max",         new SearchFun(tf_binary,   sf_MAX)));    
    
    mapFunsNoArg.insert(std::make_pair("exif.make",     new SearchFun(tf_push_str,  sf_exif_make,  1)));
    mapFunsNoArg.insert(std::make_pair("exif.model",    new SearchFun(tf_push_str,  sf_exif_model, 1)));
    mapFunsNoArg.insert(std::make_pair("exif.t",        new SearchFun(tf_push_int,  sf_exif_t)));
    mapFunsNoArg.insert(std::make_pair("exif.f",        new SearchFun(tf_push_int,  sf_exif_f)));
    mapFunsNoArg.insert(std::make_pair("exif.focalLen", new SearchFun(tf_push_int,  sf_exif_focalLen)));
    mapFunsNoArg.insert(std::make_pair("exif.iso",      new SearchFun(tf_push_int,  sf_exif_iso)));
    mapFunsNoArg.insert(std::make_pair("exif.date",     new SearchFun(tf_push_date, sf_exif_date)));
    
    mapFunsNoArg.insert(std::make_pair("jww.rename01", new SearchFun(tf_unary, sf_jww_rename01)));
  }
  
  strErr = "";
  strLog = "";
  std::string::size_type posA, posB, posC, posD, posE;
  std::string strExprOut = "";

  // todo
  //   rate.num > 5 ; rate.best > 3 : *
  //   rate.num
  
  
  // translate query to stack language
  {
    strLog = "<p>Translating search query</p><p>" + strExpr + "</p>\n";

    std::map<std::string, StrSearchGroup> mapWords;
    std::map<std::string, StrSearchGroup>::iterator itWords;
    std::vector<std::string> vConds;
    std::string strOp1, strOp2, strOp3;
    posA = 0;
    int state = 0;
    bool bFlush = false;
    while (state < 100)
    {
      // skip whitespace
      while (isblank(strExpr[posA]) && strExpr.length() > posA)
        posA++;
      
      if (strExpr.length() > posA)
      {      
        if (strExpr[posA] == ':' || strExpr[posA] == ';')
        {
          posC = posA;
          posB = posA;
        }
        else
        {
          posC = posA;
          while (isblank(strExpr[posC]) == 0 && strExpr[posC] != '"' && strExpr.length() > posC)
            posC++;
          if (isblank(strExpr[posC]))
            posC--;
          // posC now points to end of strExpr, or '"', or a blank
          if (strExpr.length() > posC && strExpr[posC] == '"')
          {
            posB = posC;
            posC = strExpr.find('"', posB+1);
            if (posC == std::string::npos)
            {
              strErr = itoStr(posB, ' ', 1) + ": expected closing \"";
              break;
            }
          }
          else
          {
            posB = posA;          
          }        
        }
        
        //std::cout << "\ntoken: " << strExpr.substr(posA, posC-posA+1) << " " << posA << " " << posB << " " << posC <<"\n";
        
        //std::cout << "state=" << state << "\n";
        switch (state)
        {
          case 0:
            if (strExpr[posC] == '"')
            {
              std::string strFlags = strExpr.substr(posA, posB-posA);
              //std::cout << "search word (" << strFlags << ") \"" << strExpr.substr(posB+1, posC-posB-1) << "\"\n";
              
              bool bNot = strFlags.length() > 0 && strFlags[0] == '!';
              if (bNot)
                strFlags = strFlags.substr(1);
              
              itWords = mapWords.find(strFlags);
              if (itWords == mapWords.end())
              {
                mapWords.insert(std::make_pair(strFlags, StrSearchGroup(strFlags)));
                itWords = mapWords.find(strFlags);
              }
              itWords->second.Add(strExpr.substr(posB+1, posC-posB-1), bNot);
              
              posA = posC+1;
            }
            else if (strExpr[posC] == ':' || strExpr[posC] == ';')
            {
              bFlush = true;
              // 
              if (strExpr[posC] == ':')
                state = 50;
              posA++;
            }
            else
            {
              // first operand
              strOp1 = strExpr.substr(posA, posC-posA+1);
              posA = posC+1;
              state = 1;
            }
            break;
            
          case 1:
            strOp2 = strExpr.substr(posA, posC-posA+1);
            posA = posC+1;
            state = 2;
            break;
            
          case 2:
            strOp3 = strExpr.substr(posA, posC-posA+1);
            posA = posC+1;
            state = 0;
            vConds.push_back(strOp1 + " " + strOp3 + " " + strOp2);
            //std::cout << "cond: " << strOp1 << " " << strOp2 << " " << strOp3 << "\n";
            break;
            
          case 50:
            // append until ';' not inside of '"'
            bool bInStr = false;
            strExprOut += " ";
            while ((strExpr[posA] != ';' || bInStr == true) && posA < strExpr.length())
            {
              if (strExpr[posA] == '"')
                bInStr = !bInStr;
              strExprOut += strExpr[posA];
              posA++;
            }
            if (strExpr[posA] == ';')
              posA++;
            state = 0;
            break;          
        }
      }
      
      if (bFlush || posA >= strExpr.length()-1)
      {
        if (state == 1 || state == 2)
        {
          strErr = "incomplete condition: " + strOp1;
          if (state == 2)
            strErr += " " + strOp2;
          break;
        }
                
        int nParts = -1;
        if (mapWords.size())
        {
          for (itWords = mapWords.begin(); itWords != mapWords.end(); itWords++)
          {
            itWords->second.AddCond(strExprOut);
            nParts++;
          }
        }

        if (vConds.size())
          nParts++;
        for (unsigned int n=0; n<vConds.size(); n++)
        {
          strExprOut += " " + vConds[n];
          if (n > 0)
            strExprOut += " *";
        }
                        
        for (; nParts>0; nParts--)
          strExprOut += " *";
        
        mapWords.clear();
        vConds.clear();
        bFlush = false;
      }
      
      if (posA >= strExpr.length() - 1)
        state = 100;      
    }    
  }
  
  strExpr = strExprOut;
  strLog += "<p>" + strExpr + "</p>\n";
//  std::cout << "compiled to:\n" << strExpr << "\n";
  
  // compile expression into vInstr
  std::string strT, strArg;
  std::map<std::string, SearchFun*>::iterator it;
  TestStack teststack;
  int nStackMax = 0;
  int nTempStrings = 0;
  
  fNeedsPCom = false;
  fNeedsPKey = false;
  
  if (strErr.length() == 0)
  {
    strLog += "<p>" + strExpr + "</p><p></p>\n";
    
    posA = 0;
    while (strExpr.length() > posA)
    {
      while (strExpr.length() > posA && isblank(strExpr[posA]))
        posA++;
      if (strExpr.length() == posA)
        break;
      // Jetzt zeigt posA auf den Anfang eines Ausdrucks
      posB = posA + 1;
      // Das Ende des Ausdrucks ist 
      //   - Stringende
      //   - Leerzeichen (es sei denn, vorher kommt ein ")
      //   - schließende Klammer
      posC = std::string::npos;
      posD = std::string::npos;
      posE = std::string::npos;
      {
        int state = 0;
        while (state != 99 && posB < strExpr.length())
        {
          switch (state)
          {
            case 0:
              if (isblank(strExpr[posB]))
                state = 99;
              else if (strExpr[posB] == '(')
              {
                state = 1;
                posC = posB;
                posB++;
              }
              else
                posB++;
              break;
              
            case 1:
              if (strExpr[posB] == '"')
              {
                state = 2;
                posD = posB;
              }
              posB++;
              break;
              
            case 2:
              if (strExpr[posB] == '"')
              {
                state = 3;
                posE = posB;
              }
              posB++;
              break;
              
            case 3:
              if (strExpr[posB] == ')')
              {
                state = 99;
              }
              posB++;
              break;
          }
        }
      }
      
      // posB steht hinter dem Ausdruck -- egal ob die Zeile zuende ist oder nicht.
      strT = strExpr.substr(posA, posB-posA);
      
      if (posC != std::string::npos && posD != std::string::npos && posE != std::string::npos)
      {
        strArg = strExpr.substr(posD+1, posE-posD-1);
        strT   = strExpr.substr(posA, posC-posA);
        it = mapFunsStrArg.find(strT);
        if (it != mapFunsStrArg.end())
        {
          vStringsConst.push_back(new std::string(strArg));
          vInstr.push_back(new ProgInstr(it->second, (intptr_t)(vStringsConst[vStringsConst.size()-1]) ));
          nTempStrings += it->second->nCreatesTempStrings;
          it->second->pTestFun(&teststack, strT);
          //std::cout << "Instr: " << strT << " (" << it->second->nStackDiff << "), \"" << strArg << "\"\n";
        }
        else
        {
          strErr = "unknown command: " + strT;
          break;
        }            
      }
      else
      {
        it = mapFunsNoArg.find(strT);
        if (it != mapFunsNoArg.end())
        {
          vInstr.push_back(new ProgInstr(it->second, 0));
          nTempStrings += it->second->nCreatesTempStrings;
          //std::cout << "Instr: " << strT << " (" << it->second->nStackDiff << "), 0\n";
          it->second->pTestFun(&teststack, strT);
          
          if (strT.compare("pcom.l") == 0)
            fNeedsPCom = true;
          if (strT.compare("pkey.l") == 0)
            fNeedsPKey = true;
        }
        else
        {
          for (posE=0; posE<strT.length(); posE++)
            if ((strT[posE] < '0' || strT[posE] > '9') &&
                strT[posE] != ':' &&
                strT[posE] != '-')
            {
              strErr = "unknown command: " + strT;
              break;
            }
          if (strErr.length())
            break;
          
          // it should be a constant:
          //    int as decimal
          //    date: yyyy-mm-dd
          //    time: hh:mm:ss
          intptr_t val = 0;
          JWWDateTime dt;
          if (strT.find(':') != std::string::npos)
          {
            dt.setTime(strT.c_str());
            val = dt.nTime;
            // std::cout << "Instr: push constant value (1), " << dt.getTimeAsString() << "\n";
            vInstr.push_back(new ProgInstr(new SearchFun(tf_push_time, sf_CONST), val));
            tf_push_time(&teststack, strT);
          }
          else if (strT.find('-') != std::string::npos)
          {
            dt.setDate(strT.c_str());
            val = dt.nDate;
            //std::cout << "Instr: push constant value (1), " << dt.getDateAsString() << "\n";
            vInstr.push_back(new ProgInstr(new SearchFun(tf_push_date, sf_CONST), val));
            tf_push_date(&teststack, strT);
          }
          else
          {
            val = atoi(strT.c_str());
            //std::cout << "Instr: push constant value (1), " << val << "\n";
            vInstr.push_back(new ProgInstr(new SearchFun(tf_push_int, sf_CONST), val));
            tf_push_int(&teststack, strT);
          }
        }
      }
      
      //
      posA = posB;
      if (nStackMax < teststack.Size())
        nStackMax = teststack.Size();
      
      if (teststack.GetErr().length())
      {
        strErr = teststack.GetErr();
        break;
      }
      else
      {
        strLog.append("Stack after instruction \"" + strT + "\":<br/>\n");
        teststack.Print(strLog);
      }
    }
  }
  
  if (strErr.length() == 0)
  {
    if (teststack.Size() != 1)
      strErr = "resulting stacksize must be 1, but is " + itoStr(teststack.Size(), ' ', 1);
    else if (teststack.GetType(0) != TestStack::eInt)
      strErr = "last element on stack must be an int";
  }

  if (strErr.length())
  {
    std::cerr << "strErr=" << strErr << "\n";
    stack = 0;
  }
  else
  {
    stack = new intptr_t[nStackMax+1];
    context.vStringsTemp.resize(nTempStrings);
  }
}

SearchProg::~SearchProg()
{
  if (stack)
    delete stack;
  
  unsigned int size = vInstr.size();
  for (unsigned int n=0; n<size; n++)
    delete vInstr[n];
  
  size = vStringsConst.size();
  for (unsigned int n=0; n<size; n++)
    delete vStringsConst[n];
}

int SearchProg::Run(ItPicInGroup* pPic, ItGroup* pParent, std::string strBasePath, std::string strPCom, std::string strPKey)
{
  if (stack == 0) // catch compilation error
    return 0;
  
  bool fDebug = false; // setting this to "true" results in slow searches!
  if (fDebug)
    std::cout << pPic->pPic->GetPath() << ":";
  intptr_t* pTopOfStack = stack;
  context.pPic        = pPic;
  context.pParent     = pParent;
  context.strBasePath = strBasePath;
  context.nTmpStrIdx  = 0;
  context.pExif       = 0;
  
  if (NeedsPCom())
    context.strPCom  = strPCom + " " + UTF8::to_lower(pPic->pPic->GetComment());
  if (NeedsPKey())
    context.strPKey  = strPKey + UTF8::to_lower(pPic->pPic->GetKeywords());
  
  unsigned int size = vInstr.size();
  for (unsigned int n=0; n<size; n++)
  {
    ptrFunSearch pFun = vInstr[n]->pFun;
    int nRet = pFun(pTopOfStack, &context, vInstr[n]->arg);
    
    if (fDebug)
      std::cout << " " << *(pTopOfStack-1);
    switch (nRet)
    {
      case 1:   
        if (fDebug)
          std::cout << " : 1\n";
        if (context.pExif)
          delete context.pExif;
        return 1;
      case -1:
        if (fDebug)
          std::cout << " : 0\n";
        if (context.pExif)
          delete context.pExif;
        return 0;
      default:
        break;
    }
  }
  
  if (context.pExif)
    delete context.pExif;
  
  if (fDebug)
    std::cout << " : " << *(pTopOfStack-1) << "\n";
  return *(pTopOfStack-1);
}

