/*
 * Copyright (C) 2005-2008, 2010, 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#ifndef TREEVIEW_H
# define TREEVIEW_H

# include <string>
# include <map>
# include <vector>

# include "perm.h"
# include "pic.h"


/**
 * Pointer to callback function for selecting a user id
 */
typedef void(*ptrFunUserIDCallback)(void*);

class TreeKwHandler;
class ItPicInGroup;
class ItGroup;
class ItGSearch;
class ItGRoot;
class User;

class TreeView
{
  friend class TreeKwHandler;
public:
  
  /**
   * Load errors: these codes may be or'ed together!
   */
  enum eLoadError { eLE_NO_ERROR=0, eLE_LINK=1, eLE_BASEPATH=2, eLE_XML=4, eLE_NOFILE=8, eLE_VERSION=16, eLE_NotUTF8=32, eLE_UTFConv=64 };
  
  /**
   * 
   */
  enum eThumbDir { eSrc=0, eHome=1, eSelect=2 };
  
  
public:
  
  /** */
  TreeView();
  
  /** */
  ~TreeView();
  
  /**
   * Is there anything except default data?
   */
  bool hasData();
  
  /**
   * Makes sure parents are expanded so <code>item</code> can be seen.
   * Returns true if one or more parents had to be expanded.
   */
  bool makeVisible(ItGroup* item);
  
  /** 
   * Loads data from a file into @ref pGRoot. @ref projectfilePath gets updated.
   * If @ref pGRoot already contained data, it will be deleted.
   * Return codes: see eLE...
   */
  int loadFromFile(std::string newProjectfilePath, std::string strCharset,
                   ptrFunUserIDCallback funptr_UserIDCallback);
  
  /**
   * Save contents of project file
   */
  SimpleXMLTransfer* saveToSXT(std::string newProjectfilePath, int nVersion);
  
  /**
   * 
   */
  int saveToProjectFile();
  
  /**
   * Save myself as a new project. If I am a view of some other data,
   * I have to transform to real data before! Doing this will not be easy
   * if there already are views of myself. Issue an error in this case (return -1)
   * and don't do anything.
   */
  int saveAsNewProject(std::string newProjectfilePath);
    
  /**
   * 
   */
  std::string resolveExternalLink(std::string linktext,
                                  ItGroup*    pGroup);
  
  /**
   * Implementation without recursive function calls.
   */
  ItGroup* browserItemPrev(ItGroup* p);
  
  /**
   * Implementation without recursive function calls.
   */
  ItGroup* browserItemNext(ItGroup* p);
        
  /**
   * 
   */
  std::string getPicPath(ItPic* item, int nSize=0);
       
  /**
   * 
   */
  std::string getProjectfilePath();
  
  /**
   * 
   */
  int addPicsFromDir(const char*            picdir, 
                     ItGroup*               pDst,
                     ItPic::eDateTimeSource eDTS);
  
  /**
   * 
   */
  int addFile(const char*            picdir, 
              ItGroup*               pDst,
              ItPic::eDateTimeSource eDTS);
  
  void Delete(ItPic* pP);
  void Delete(ItGroup* pG);
  
  /**
   * picdir is the directory of a newly added picture.
   */
  int updateBasepath(const char* picdir);
           
  /**
   * 
   */
  std::string getBasepath();
  
  /**
   * See @ref fAddDelPicAllowed
   */
  bool isAddDelAllowed();
    
  /**
   * 
   */
  bool dataSaved();
    
  /**
   * To exchange data with album settings dialog.
   */
  void exchangeThumbSettings(eThumbDir& th, std::string& tds, bool fWrite);

  /**
   * Expands a list recursively to a set (no duplicates).
   */
//todo  static void RecursiveList(const std::vector<TreeItem*>& list, std::set<TreeItem*>& set);
  
  /**
   * returns path of most recent addition
   */
  std::string GetAddPath() { return(strAddPath); };
  
private:
  bool Load_v09AItem(SimpleXMLTransfer*          xData, 
                     std::map<int, std::string>& mapIDOldNew,
                     std::map<int, int>&         mapIDOldMultiLink);
  void Load_v09ARec(SimpleXMLTransfer*          xData, 
                    std::map<int, std::string>& mapIDOldNew,
                    std::map<int, int>&         mapIDOldMultiLink);
  
  ItPicInGroup* Load_v09BItem(SimpleXMLTransfer*             xData, 
                              std::map<int, std::string>&    mapIDOldNew,
                              std::map<int, int>&            mapIDOldMultiLink,
                              std::map<std::string, ItPic*>& mapPicsPath);
  void Load_v09BRec(SimpleXMLTransfer*          xData, 
                    std::map<int, std::string>& mapIDOldNew,
                    std::map<int, int>&         mapIDOldMultiLink,
                    std::map<std::string, ItPic*>& mapPicsPath,
                    ItGroup* pParent);

  int Load_v09(std::string newProjectfilePath, SimpleXMLTransfer* data, std::string strCharset);
  int Load_v10(std::string newProjectfilePath, SimpleXMLTransfer* data,
               ptrFunUserIDCallback funptr_UserIDCallback);
  
public:
  
  /** my data  */
  ItGRoot* pGRoot;
  ItGRoot* pGOrphanedPics;
  ItGroup* GetUserRoot();

  /** for keyword handling */
  TreeKwHandler* kw;
  
  /** available thumbnail sizes */
  static const unsigned int uThumbSizes[];

  T_Perm* perm;
  User* pUser;
  
  void AddPic(ItPic* pPic);
  void AddGroup(ItGroup* pGroup);
  
  ItGroup* GetGroupPtr(std::string strID);
  ItPic*   GetPicPtr(std::string strID);
  ItPicInGroup* GetPicInGroupPtr(std::string strID, ItGroup*& pParent);
  
  void CreateOrphanedGroup();
    
private:

  std::vector<ItPic*> vPics;
  std::map<std::string, ItPic*> mapPics;
  
  std::vector<ItGroup*> vGroups;
  std::map<std::string, ItGroup*> mapGroups;
  
  std::vector<ItGSearch*> vSearches;
  
  /** The first path is special. */
  std::vector<std::string> basePaths;
    
  /** the basepath in use */
  std::string strBasePath;
  
  /**
   * Depends on basepath in use.
   * Does not refer to add and del only, but also to actions on the pictures 
   * themselves, like rotating and editing.
   */
  bool fAddDelPicAllowed;
  
  /** 
   * Views created by me. I have to stay alive as long as there
   * are some of them, as they might link back to my data.
   */
  std::vector<TreeView*> views;
  
  /**
   * path to project file
   */
  std::string projectfilePath;
  
  /**
   * Where to put thumbnails? This should not be read directly,
   * as it depends on @ref source, too. Use @ref getPicPath in any case.
   */
  eThumbDir thumbdir;
  
  /**
   * Valid in case of thumbdir == eSelect. This should not be read directly,
   * as it depends on @ref source, too. Use @ref getPicPath in any case.
   */
  std::string thumbdirselect;
  
  std::string strBasepathRelOriginal;
  
  std::string strAddPath;
  
  int nVersionInfo;
};

#endif
