/*
 * Copyright (C) 2011  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "cache.h"
#include "../shared/lib_conversions.h"
#include "../shared/hirestime.h"

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <fstream>

#ifdef WIN32
# include <windows.h>
#endif

#ifdef WIN32_FLTK13
# include <FL/fl_utf8.h>
#endif

bool CachedPic::fits(std::string npath, int nw, int nh, int nQual)
{
  return(nw == wmax && 
         nh == hmax &&
         strPathSrc.compare(npath) == 0 &&
         this->nQual == nQual);
}

CachedPic::~CachedPic()
  {
    if (FileTools::rm(strPathCache))
      std::cerr << "error deleting " << strPathCache << "\n";
  }

void CachedPic::Save(Fl_RGB_Image* pImg)
{
  HiResTime hrt;
  hrt.Capture();
  
  // save data to file
  pic_d = pImg->d();
  pic_w = pImg->w();
  pic_h = pImg->h();
  size = pic_w * pic_h * pic_d;
  
  FILE* out;
  #ifdef WIN32_FLTK13
    out = fl_fopen(strPathCache.c_str(), "wb");
  #else
    out = fopen(strPathCache.c_str(), "wb");
  #endif
  
  if (out)
  {
    //const char* buf = (const char*)(pImg->array);   todo: warum geht das nicht?
    const char* buf = (const char*)(pImg->data()[0]);
    fwrite(buf, 1, size, out);
    fclose(out);    
  }
  else    
  {
    size = 0;
    std::cout << "error opening " << strPathCache << "\n";
  }    
  std::cout << "timing.pic.cache.save=" << hrt.GetMSecSinceCapture() << "\n";   
}

Fl_RGB_Image* CachedPic::Load()
{
  HiResTime hrt;
  hrt.Capture();
  
  FILE* in;
  #ifdef WIN32_FLTK13
    in = fl_fopen(strPathCache.c_str(), "rb");
  #else
    in = fopen(strPathCache.c_str(), "rb");
  #endif
  if (in)
  {    
    uchar* new_array = new uchar[size];
    Fl_RGB_Image* new_image = new Fl_RGB_Image(new_array, pic_w, pic_h, pic_d);
    new_image->alloc_array = 1;
    
    fread((char*)(new_array), 1, size, in);
    fclose(in);
    
    std::cout << "timing.pic.cache.load=" << hrt.GetMSecSinceCapture() << "\n";   
    
    return(new_image);
  }
  else
    std::cout << "error opening " << strPathCache << "\n";    
  return(0);
}

PicCache::PicCache(int nCacheBytes)
{
  nName = 0;
  this->nCacheBytes = nCacheBytes;
  
  // create a unique directory name
#ifndef WIN32
  // man 3 mkdtemp
  char acBuf[5+7+6+1] = "/tmp/facacheXXXXXX";
  if (mkdtemp(acBuf) == NULL)
  {
    // simply don't use the cache:
    this->strCacheDir = "";
  }
  else
  {
    this->strCacheDir = acBuf;
  }
#else
  this->strCacheDir = TemporaryTextFile::GetGlobalTempDirPath() + "facache" + itoHexStr(GetCurrentProcessId() );
#endif
  
  if (!this->strCacheDir.empty())
  {
    FileTools::makeSureDirExists(this->strCacheDir);
  }
}

PicCache::~PicCache()
{
  // delete all cache files
  for (std::list<CachedPic*>::iterator it=liste.begin(); it != liste.end(); it++)
    delete *it;
  
  // delete directory
#ifndef WIN32
  remove(strCacheDir.c_str());
#else  
# ifdef WIN32_FLTK13
  fl_rmdir(strCacheDir.c_str());
# endif
#endif
}

CachedPic* PicCache::Request(std::string strPath, int w, int h, int nQual)
{
  for (std::list<CachedPic*>::iterator it=liste.begin(); it != liste.end(); it++)
    if ((*it)->fits(strPath, w, h, nQual))
    {
      // move item to front
      liste.splice(liste.begin(), liste, it, it);
      
      // up-to-date?
      if (FileTools::isFileUpToDate(strPath, (*it)->strPathCache) < 1)
      {
        // Cache item reset, because it is not up-to-date
        (*it)->size = 0;
      }
      return *it;
    }
  
  // item is not in cache
  CachedPic* tmp = new CachedPic(strPath, strCacheDir + "/" + itoHexStr(nName++, 4), w, h, nQual);
  liste.push_front(tmp);
  
  // Size limit?
  int nSum = 0;
  for (std::list<CachedPic*>::iterator it=liste.begin(); it!=liste.end(); it++)
    nSum += (*it)->size;
  while (nSum > nCacheBytes)
  {
    std::list<CachedPic*>::reverse_iterator it=liste.rbegin();
    nSum -= (*it)->size;
    delete *it;
    liste.pop_back();
  }  
  
  return(tmp);
}

