/*
 * Copyright (C) 2005,2008,2011-2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "picbox.h"

#include <iostream>
#include <cstring>
#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/Fl_PNG_Image.H>
#include "../config.h"
#include "../base/pic.h"
#include "../base/group.h"
#include "../base/treeview.h"
#include "../base/workerdata.h"
#include "../base/basetools.h"
#include "../shared/filetools.h"
#include "../shared/hirestime.h"
#include "../shared/thread.h"
#include "../shared/readjpeg.h"
#include "../shared/imgxform.h"
#include "../shared/lib_conversions.h"
#include "gui.h"

#ifdef WIN32
# include <windows.h>
#else
# include <unistd.h>
#endif

#ifdef USE_LIBGRAPHICSMAGICK
# include <magick/api.h>
#endif

PicCache* PicBox::pPicCache = 0;

bool GetShrinkInfo(int& nWNew, int& nHNew, int nWMax, int nHMax)
{
  // need to scale down?
  if (nWNew > nWMax || nHNew > nHMax)
  {
    if ((float)nHNew / nHMax > (float)nWNew / nWMax)
    {
      nWNew = (nWNew * nHMax)/nHNew;
      nHNew = nHMax;
    }
    else
    {
      nHNew = (nHNew * nWMax)/nWNew;
      nWNew = nWMax;
    }
    return(true);
  }
  else
    return(false);
}

void* LoadAndScalePic(void* wd)
{
  LoadAndScaleData* lasd = (LoadAndScaleData*)(wd);
    
  lasd->errstring = "";
  lasd->nRetCode  = 0;
  
  bool fJPEG, fMovie;
  std::string strSuff;
  {
    strSuff = strL(FileTools::getSuffix(lasd->pCachedPic->strPathSrc));
    fJPEG = (strSuff.compare(".jpg") == 0 || strSuff.compare(".jpeg") == 0);
    fMovie = BaseTools::SuffixIsMovie(strSuff);
  }
  
  if (!FileTools::fileExists(lasd->pCachedPic->strPathSrc))
  {
    lasd->errstring  = "Error loading image: no such file: ";
    lasd->errstring += FileTools::name(lasd->newpic->GetPath());
    lasd->nRetCode = -1;
  }
  else if (fMovie)
  {
    lasd->errstring = FileTools::name(lasd->newpic->GetPath());
    lasd->errstring += ": please view movies using a plugin.";
    lasd->nRetCode = -1;
  }
  else
  {
    HiResTime hrt;
    
    // load image file
    uchar* array = 0;
    int nW, nH, nD;
    {
      hrt.Capture();
      
      if (fJPEG && USE_LIBJPEG != 0)
      {
        int nDummy;
        nW = lasd->pCachedPic->wmax;
        nH = lasd->pCachedPic->hmax;
        array = LoadScaledJPEG(lasd->pCachedPic->strPathSrc.c_str(), nW, nH, nDummy);
        nD = 3;
      }
      else if (strSuff.compare(".png") == 0)
      {
        Fl_RGB_Image* img = new Fl_PNG_Image(lasd->pCachedPic->strPathSrc.c_str());
                
        array = (uchar*)(img->data()[0]);
        nW = img->w();
        nH = img->h();
        nD = img->d();
        if (img->ld() != 0 && img->ld() != nW*nD)
        {
          std::cout << "Problem: " << img->ld() << "\n";std::cout.flush();
        }
        img->alloc_array = 0;
        delete img;
      }       
      
      if (array == 0)
      {
#ifdef USE_LIBGRAPHICSMAGICK
        // todo: check file type/extension before doing long operations?
        //       LIBGRAPHICSMAGICK takes a lot of time until it knows it 
        //       can't read an avi file.
        std::cout << "Using libgraphicsmagick to load file\n";
        
        ExceptionInfo exception;
        Image* image = 0;
        ImageInfo* image_info;        
        GetExceptionInfo(&exception);
        
        image_info=CloneImageInfo((ImageInfo *) NULL);    
        strcpy(image_info->filename, lasd->pCachedPic->strPathSrc.c_str());
        image = ReadImage(image_info, &exception);
        CatchException(&exception);
        DestroyImageInfo(image_info);
        if (image != 0)
        {
          nD = 3;
          nW = image->columns;
          nH = image->rows;
          array = new uchar[nH*nW*nD];
          DispatchImage(image, 0, 0, nW, nH, "RGB", CharPixel, array, &exception);
          DestroyImage(image);
          DestroyExceptionInfo(&exception);
        }
#endif
      }
      
      if (array == 0)
      {
        lasd->errstring  = "Error loading image: ";
        lasd->errstring += FileTools::name(lasd->newpic->GetPath());
        lasd->nRetCode = -1;
        lasd->fActive  = false;
        if (lasd->fThread)
          SimpleThread::Exit();
        return((void*)(lasd->nRetCode));
      }
      
      std::cout << "timing.pic.load=" << hrt.GetMSecSinceCapture() << " " << lasd->pCachedPic->strPathSrc << "\n";            
    }

    // scale image, set (intermediate) output
    {      
      int nWNew = nW;
      int nHNew = nH;
      
      uchar* array_scaled = array;
      
      if (GetShrinkInfo(nWNew, nHNew, lasd->pCachedPic->wmax, lasd->pCachedPic->hmax))
      {
        hrt.Capture();
        array_scaled = ImgXForm::Sample(array, nW, nH, nD, nWNew, nHNew);
        std::cout << "timing.pic.sample=" << hrt.GetMSecSinceCapture() << "\n";
      
        if (lasd->nQualOverSpeed > 0)
        {
          lasd->res_fast = new Fl_RGB_Image(array_scaled, nWNew, nHNew, nD);
          lasd->res_fast->alloc_array = 1;
          if (lasd->pPicBox)
            lasd->pPicBox->SetFastImg(lasd->res_fast);
          
          if (nW >= nWNew*2 && nH >= nHNew*2) // both conditions are needed if (aspectratio != 1)
          {
            float flDiv  = (1.0*nW)/nWNew;
            float flDivH = (1.0*nH)/nHNew;
            if (flDiv > flDivH)
              flDiv = flDivH;
            
            uchar* tmp = array;
            hrt.Capture();
            array = ImgXForm::Shrink(tmp, nW, nH, nD, floor(flDiv));
            std::cout << "timing.pic.shrink=" << hrt.GetMSecSinceCapture() << "\n";
            delete[] tmp;
          }
          
          hrt.Capture();
          array_scaled = ImgXForm::Bilinear(array, nW, nH, nD, nWNew, nHNew);
          std::cout << "timing.pic.bilinear=" << hrt.GetMSecSinceCapture() << "\n";
        }
        delete[] array;
      }
      
      lasd->res = new Fl_RGB_Image(array_scaled, nWNew, nHNew, nD);
      lasd->res->alloc_array = 1;                        
    }    
  }
  
  lasd->fActive  = false;
  if (lasd->fThread)
    SimpleThread::Exit();
  
  return((void*)(lasd->nRetCode));
}

PicBox::PicBox(int X,int Y,int W,int H,const char *l)
  : Fl_Box(X,Y,W,H,l)
{
  textbox  = new TextBox(X, Y, W, H);
  nVerbose = 3;
  box(FL_FLAT_BOX);
  
  pFastImg = 0;
  myRGBImg = 0;
  if (pPicCache == 0)
    pPicCache = new PicCache(200*1024*1024);
}

void PicBox::boxcallback(Fl_Callback* fnptr, void* dtptr)
{
  textbox->callback(fnptr, dtptr);
}

PicBox::~PicBox()
{
  // Wait until thread has finished, because it will access my data
  if (lasd_thread.fThread)
  {
    if (lasd_thread.fActive)
    {
      fl_cursor(FL_CURSOR_WAIT);
      while (lasd_thread.fActive)
      {
        Fl::wait(0.1);
      }
      fl_cursor(FL_CURSOR_DEFAULT);
    }
  }
  
  // delete pPicCache; must not be deleted, because another window might be open and using it.
  // It will be deleted at the end of main().
    
  Fl::delete_widget(textbox);
  
  if (pFastImg)
  {
    pFastImg->uncache();
    delete pFastImg;    
  }
  if (myRGBImg)
  {
    myRGBImg->uncache();
    delete myRGBImg;
  }
  if (lasd_thread.res_fast)
  {
    lasd_thread.res_fast->uncache();
    delete lasd_thread.res_fast;
  }
  if (lasd_thread.res)
  {
    lasd_thread.res->uncache();
    delete lasd_thread.res;
  }
}

void PicBox::toggleVerbosity()
{
  nVerbose = (++nVerbose) % ItPic::eDescrLevels;
  
  textbox->setDescr(myPic->pPic->getDescr(nVerbose, false, pTreeView->getBasepath()).c_str());
  textbox->recalcLayoutInto(x(), y(), w(), h());
  
  if (!textbox->visible())
  {
    textbox->show();
  }
  
  textbox->redraw();
  redraw();
}

int PicBox::handle(int event)
{
  int X,Y,W,H; 
  
  X = x();
  Y = y();
  W = w();
  H = h();
  
  int nRetCode = textbox->handle(event);
  
  if (nRetCode)
    return(nRetCode);
  
  switch (event) 
  {
    case FL_FOCUS:
      //    std::cout << "FL_FOCUS\n";
      return(1);
      break;
    case FL_UNFOCUS:
      //    std::cout << "FL_UNFOCUS\n";
      return(1);
      break;
      
    case FL_KEYBOARD:
    case FL_SHORTCUT:
      // std::cout << "Keypress: " << Fl::event_key() << "\n";
      if (!Fl::event_state(FL_ALT))
      {
        switch (Fl::event_key())
        {
          case 101: // todo: 'e'
            do_callback(this, PROPERTIES);
            return(1);      
            break;
            
          case 107: // todo: 'k'
            do_callback(this, ADD_KEYWORDS);
            return(1);      
            break;
            
          case 118: // todo: 'v'
            toggleVerbosity();            
            return(1);
            break;
            
          case FL_Right:
          case 32: // todo: space
            do_callback(this, PICTURE_NEXT);
            return(1);
            break;
            
          case FL_Left:
          case FL_BackSpace:
            do_callback(this, PICTURE_PREV);
            return(1);
            break;
            
          case 116: // todo: 't'
            do_callback(this, TOGGLE_MARK);
            return(1);
            break;
            
          case FL_Escape:
            do_callback(this, GROUP);
            return(1);
            break;
            
          case 120: // todo: 'x'
            do_callback(this, (int)(RATE)+1);
            return(1);
            break;
            
          default:
            int key = Fl::event_key();
            if (key >= 48 && key <= 53)
            {
              do_callback(this, (int)(RATE)+key-48);
              return 1;
            }
            else
              std::cout << "Keypress: " << Fl::event_key() << "\n";
            break;
        }
      }
      break;
      
    case FL_PUSH:
      
      // this is to get the focus after pushing a mouse button
      if (handle(FL_FOCUS))
        Fl::focus(this);
      
      return 1;
      break;
      
    case FL_RELEASE:
      
      if (Fl::event_button() == FL_RIGHT_MOUSE)
      {
        //      std::cout << "picbox context menu\n";
        nMX = Fl::event_x();
        nMY = Fl::event_y();
        do_callback(this, CONTEXT);
      }
      else if (Fl::event_x() > X + (W>>1))
      {
        if (Fl::event_y() > Y + (H>>1))
        {
          //        std::cout << "next picture\n";
          do_callback(this, PICTURE_NEXT);
        }
        else
        {
          //        std::cout << "group\n";
          do_callback(this, GROUP);
        }
      }
      else
      {
        if (Fl::event_y() > Y + (H>>1))
          do_callback(this, PICTURE_PREV);
        else
          toggleVerbosity();
      }    
      return(1);        
  }
  
  return Fl_Box::handle(event);
}

void PicBox::resize(int x, int y, int w, int h)
{
  //  std::cout << "void PicBox::resize(" << x << ", " << y << ", " << w << ", " << h << ")\n";
  textbox->recalcLayoutInto(x, y, w, h);
  Fl_Box::resize(x, y, w, h);
}

void PicBox::draw()
{
  //  std::cout << "void PicBox::draw()\n";
  Fl_Box::draw(); 
  if (textbox->visible())
  {
    textbox->fMarked = myPic->fMarked;
    textbox->draw();
  }
}

void PicBox::setDescr(std::string newdescr)
{
  textbox->setDescr(newdescr.c_str());
  textbox->labelsize(labelsize());
  
  textbox->recalcLayoutInto(x(), y(), w(), h());  
}  

void PicBox::UpdateTextBox()
{
  setDescr(myPic->pPic->getDescr(nVerbose, false, pTreeView->getBasepath()).c_str());
  redrawTextBox();
}  

void PicBox::redrawTextBox()
{
  //  std::cout << "void PicBox::redrawTextBox()\n";
  textbox->redraw();
  redraw();
}  

int PicBox::setPic(ItPicInGroup*  newpic,
                   TreeView*      pTreeView,
                   T_optionen*    opts,
                   UserInterface* pGUI)
{
  std::string basepath = pTreeView->getBasepath();
  
  int nQualOverSpeed = opts->pool->getInt("display.QualOverSpeed", 1);
 
  this->pTreeView = pTreeView;
  this->pGUI = pGUI;
  myPic = newpic;
  setDescr(myPic->pPic->getDescr(nVerbose, false, basepath));

  CachedPic* pFromCache = pPicCache->Request(basepath + newpic->pPic->GetPath(), w(), h(), nQualOverSpeed);
      
  if (lasd_thread.fThread)
  {
    // Nur auf einen laufenden Thread warten, wenn man dessen Ergebnis braucht
    if (lasd_thread.pCachedPic == pFromCache)
    {
      if (lasd_thread.fActive)
      {
        fl_cursor(FL_CURSOR_WAIT);
        while (lasd_thread.fActive)
        {
          // Gibt es schon eine schnell berechnete Vorversion des Bilds?
          if (lasd_thread.res_fast)
          {
            SetFastImg(lasd_thread.res_fast);
            lasd_thread.res_fast = 0;
          }
          Fl::wait(0.1);
        }
        fl_cursor(FL_CURSOR_DEFAULT);
      }    
      
      if (lasd_thread.nRetCode == 0)
      {
        lasd_thread.pCachedPic->Save(lasd_thread.res);
        lasd_thread.res->uncache();
        delete lasd_thread.res;
        lasd_thread.res = 0;
        if (lasd_thread.res_fast)
        {
          lasd_thread.res_fast->uncache();
          delete lasd_thread.res_fast;
          lasd_thread.res_fast = 0;
        }
      }
      
      lasd_thread.fThread = false;
    }
    else
    {
      if (lasd_thread.fActive)
      {
        // todo: Thread-Priorität absenken
      }
      else
      {
        lasd_thread.fThread = false;
      }
    }
  }
      
  int nRetCode = 0;

    
  Fl_RGB_Image* pImg = 0;
  if (pFromCache->size == 0)
  {
    LoadAndScaleData lasd;
    lasd.pCachedPic    = pFromCache;
    lasd.newpic        = myPic->pPic;
    lasd.flAspectRatio = opts->flAspectRatio;
    lasd.fThread       = false;
    lasd.pPicBox       = this;
    lasd.nQualOverSpeed= nQualOverSpeed;
    
    fl_cursor(FL_CURSOR_WAIT);
    Fl::check();
    // load the current picture without creating a thread
    if (LoadAndScalePic((void*)(&lasd)))
    {
      // error
      image((Fl_Image*)0);
      deimage((Fl_Image*)0);
      setLabel(lasd.errstring);
      redraw();
      fl_cursor(FL_CURSOR_DEFAULT);
      return(-1);
    }
    pFromCache->Save(lasd.res);
    pImg = lasd.res;
  }
  else
  {
    pImg = pFromCache->Load();
  }
    
  label((const char*)0); // needed to have image centered
  image(pImg);
  deimage(pImg);
  redraw();
  fl_cursor(FL_CURSOR_DEFAULT);
        
  if (pFastImg)
  {
    pFastImg->uncache();
    delete pFastImg;
    pFastImg = 0;
  }
  
  if (myRGBImg)
  {
    myRGBImg->uncache();
    delete myRGBImg;
  }
  myRGBImg = pImg;
    
  
  if (lasd_thread.fThread)
  {
    // todo: Thread-Priorität wieder erhöhen
  }
  else
  {
    pFromCache = 0;
    ItPicInGroup* pGuess = PicNext();
    
    if (pGuess)
    {
      pFromCache = pPicCache->Request(basepath + pGuess->pPic->GetPath(), w(), h(), nQualOverSpeed);
      if (pFromCache->size <= 0)
      {
        lasd_thread.newpic  = pGuess->pPic;
        lasd_thread.fThread = true;
      }
    }
    
    if (lasd_thread.fThread == false)
    {
      pGuess = PicPrev();
      if (pGuess)
      {
        pFromCache = pPicCache->Request(basepath + pGuess->pPic->GetPath(), w(), h(), nQualOverSpeed);
        if (pFromCache->size <= 0)
        {
          lasd_thread.newpic  = pGuess->pPic;
          lasd_thread.fThread = true;
        }
      }
    }
    
    if (lasd_thread.fThread)
    {
      lasd_thread.pCachedPic    = pFromCache;
      lasd_thread.flAspectRatio = opts->flAspectRatio;
      lasd_thread.fActive       = true;
      lasd_thread.pPicBox       = 0;
      lasd_thread.res           = 0;
      lasd_thread.res_fast      = 0;
      lasd_thread.nQualOverSpeed= nQualOverSpeed;
            
      if (SimpleThread::Create(LoadAndScalePic, (void*)&lasd_thread))
      {
        std::cerr << "error creating thread\n";
        lasd_thread.fThread = false;
      }
    }
  }
      
  return(nRetCode);
}

void PicBox::setLabel(std::string somestring)
{
  mylabel = somestring;
  label(mylabel.c_str());
}

void PicBox::SetFastImg(Fl_RGB_Image* pImg)
{
  label((const char*)0); // needed to have image centered
  image(pImg);
  deimage(pImg);
  redraw();
  Fl::wait(0);
  pFastImg = pImg;
  lasd_thread.res_fast = 0; // to prevent double delete
}

ItPicInGroup* PicBox::PicNext()
{
  return pGUI->GetLastDisplayedGroup()->GetPicNext(myPic->pPic);
}

ItPicInGroup* PicBox::PicPrev()
{
  return pGUI->GetLastDisplayedGroup()->GetPicPrev(myPic->pPic);
}
