/*
 * Copyright (C) 2005, 2010, 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "textbox.h"

#include <iostream>
#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <cstdlib>
#include "../shared/textparser.h"


class TextTableCell
{
public:
  TextTableCell(std::string strT, int nWL, int nWT)
  {
    strText = strT;
    nWidthOneLine = nWL;
    nWidthLongestToken = nWT;
  }
  
  std::string strText;
  int nWidthOneLine;
  int nWidthLongestToken;
};

class TextTableCol
{
public:
  TextTableCol(int nWL, int nWT)
  {
    nWidthOneLineMax = nWL;
    nWidthLongestTokenMax = nWT;
    nWidth = 0;
  };
  
  void TrackMax(int nWL, int nWT)
  {
    if (nWidthOneLineMax < nWL)
      nWidthOneLineMax = nWL;
    if (nWidthLongestTokenMax < nWT)
      nWidthLongestTokenMax = nWT;
  };
  
  int nWidthOneLineMax;
  int nWidthLongestTokenMax;
  int nWidth;
};


TextBox::TextBox(int X,int Y,int W,int H,const char *l)
  : Fl_Box(X,Y,W,H,l)
{
  descr            = "default description";
  fTextLayoutDone  = false;
  fLinkCursor      = false;
  fMarked          = false;
}

TextBox::~TextBox()
{
//  std::cout << "TextBox::~TextBox()\n";
  if (fLinkCursor)  
  {
//    std::cout << "Resetting to default cursor\n";
    fl_cursor(FL_CURSOR_DEFAULT, FL_FOREGROUND_COLOR, FL_BACKGROUND_COLOR);
  }
}

int TextBox::handle(int event)
{
  int X,Y,W,H; 
  
  if (event == FL_LEAVE && fLinkCursor)
  {
    fl_cursor(FL_CURSOR_DEFAULT, FL_FOREGROUND_COLOR, FL_BACKGROUND_COLOR);
    fLinkCursor = false;
    return 1;
  }
  
  X = x();
  Y = y();
  W = w();
  H = h();
  
  if (
      !Fl::event_inside(X,Y,W,H)
      )    
  {
    //    std::cout << "Event ausserhalb textbox\n";
    return 0;
  }

  
//  std::cout << "Event in textbox\n";
  
  
  switch (event) 
  {
   case FL_ENTER:
    return 1;
    break;
   
   case FL_MOVE:
    if (links.size())
    {
      int   nH     = (int)(labelsize());
      bool  fLCNew = false;
      
      for (unsigned int n=0; n<links.size() && !fLCNew; n++)
      {
        if (
            Fl::event_inside(links[n].x,links[n].y+y(),links[n].w,nH)
            )          
        {
          fLCNew = true;
        }
      }
      if (fLCNew && !fLinkCursor)
      {
        fLinkCursor = true;
        fl_cursor(FL_CURSOR_HAND, FL_FOREGROUND_COLOR, FL_BACKGROUND_COLOR);
      }
      else if (!fLCNew && fLinkCursor)
      {
        fl_cursor(FL_CURSOR_DEFAULT, FL_FOREGROUND_COLOR, FL_BACKGROUND_COLOR);
        fLinkCursor = false;
      }
    }
    return 1;
    break;
    
   case FL_PUSH:
    return 1;
    break;
    
   case FL_RELEASE:
    int nH = (int)(labelsize());
    nClickedLink = -1;
    for (unsigned int n=0; n<links.size() && nClickedLink < 0; n++)
    {
      if (
          Fl::event_inside(links[n].x,links[n].y+y(),links[n].w,nH)
          )        
      {
//        std::cout << "Link gedrückt: " << links[n].dest << "\n";
        nClickedLink = n;
        do_callback();
      }   
    }    
    return(1);        
  }
  
  return Fl_Box::handle(event);
}

void TextBox::resize(int x, int y, int w, int h)
{
  Fl_Box::resize(x, y, w, h);
}

void TextBox::draw()
{
//  std::cout << "void TextBox::draw()\n";
  if (!fTextLayoutDone)
  {
//    std::cout << "void TextBox::draw() ruft doTextLayout(false);\n";
    doTextLayout(false);
  }
  
//  std::cout << "void TextBox::draw() ruft doTextLayout(true);\n";
  doTextLayout(true);  
}

void TextBox::setDescr(const char* newdescr)
{
  descr           = newdescr;
  fTextLayoutDone = false;
}

void TextBox::recalcLayoutInto(int x, int y, int w, int h)
{
//  std::cout << "void TextBox::recalcLayoutInto(int x, int y, int w, int h)\n";
  doTextLayout(false, w);
  resize(x, y, uTextWidth+2*eRAND, uTextHeight+2*eRAND);
}

unsigned int TextBox::LayoutPrint(bool fPrint,
                                  Fl_Color& textcolor,
                                  std::string   strText,
                                  int  nX0,     int nY0,
                                  int  nW,      int nWMax,
                                  unsigned int& uWidth)
{
  int X = nX0;
  int Y = nY0;
  unsigned int uTokenW;
  TextParser   p;
  std::string  token;
  bool         fNL;
  bool         fParSep;
  int          nNLEnd  = 0;
  unsigned int uLineW  = 0;
  unsigned int uSpaceW = (int)(fl_width(" "));
  bool fLineUsed = false;
  
  TTextBoxLinkData link;
  
  p.setText(strText.c_str());

  // table data
  int nCellsPerRow = -1;
  std::vector<TextTableCell> table;
  std::vector<TextTableCol>  cols;
  int nCellWidthMin = (int)(fl_width("Wl"));
  unsigned int uTableWidth = 0;
    
  while (p.hasToken())
  {
    p.getToken(token, fNL);
    
    nCellsPerRow = -1;
    if (token.find("<table") == 0)
    {
      table.clear();
      cols.clear();
      
      // Read cell contents
      {
        bool fCell = false;
        std::string strCell = "";
        int nWidthLongestToken = 0;
        int nWidth = 0;          
        int nCellMul = 1; // Number of cells with same content
        
        while (token.find(">") == std::string::npos)
        {
          p.getToken(token, fNL);
        }

        while (token.compare("</table>") != 0)
        {
          p.getToken(token, fNL);
          if (fCell)
          {
            if (token.compare("</td>") == 0 || token.compare("</th>") == 0)
            {
              for (int n=0; n<nCellMul; n++)
                table.push_back(TextTableCell(strCell, nWidth, nWidthLongestToken));
              fCell = false;
            }
            else
            {
              int nW = (int)(fl_width( token.c_str() ));
              if (nWidthLongestToken < nW)
                nWidthLongestToken = nW;
              if (strCell.length())
              {
                strCell += " ";
                nWidth += uSpaceW;
              }
              strCell += token;
              nWidth += nW;
            }
          }
          else
          {
            if (token.compare("</tr>") == 0 && nCellsPerRow < 0) 
              nCellsPerRow = table.size();
            
            if (token.find("<td") == 0 || token.find("<th") == 0)
            {
              fCell = true;
              strCell = "";
              nCellMul = 1;
              nWidthLongestToken = 0;
              nWidth = 0;          
              
              do
              {
                std::string::size_type pos = token.find("colspan=\"");
                if (pos != std::string::npos)
                {
                  std::string strNum = token.substr(pos+9, token.find("\"", pos+9)-pos-9);
                  nCellMul = atoi(strNum.c_str());
                }
                if (token.find(">") == std::string::npos)
                  p.getToken(token, fNL);
              }
              while (token.find(">") == std::string::npos);
            }
          }
        }
      }
      
      // determine basic layout data
      {
        int nCol = 0;
        for (int n=0; n<nCellsPerRow; n++)
          cols.push_back(TextTableCol(table[n].nWidthOneLine, table[n].nWidthLongestToken));
        for (unsigned int n=nCellsPerRow; n<table.size(); n++)
        {
          cols[nCol].TrackMax(table[n].nWidthOneLine, table[n].nWidthLongestToken);
          if (++nCol == nCellsPerRow)
            nCol = 0;
        }
      }
      
      // set col widths
      {
        // Are there small rows?
        // Sum size of big rows.
        int nRemainingTableWidth = nW - 2 * eRAND - (cols.size()-1)*nCellWidthMin;
        int nBigSum = 0;
        for (unsigned int n=0; n<cols.size(); n++)
        {
          if (cols[n].nWidthOneLineMax == cols[n].nWidthLongestTokenMax ||
              cols[n].nWidthOneLineMax <= nCellWidthMin)
          {
            cols[n].nWidth = cols[n].nWidthOneLineMax;
            nRemainingTableWidth -= cols[n].nWidthOneLineMax;
          }
          else
            nBigSum += cols[n].nWidthOneLineMax;
        }
        
        // set big rows
        float flFakt = (1.0*nRemainingTableWidth) / nBigSum;
        if (flFakt > 1.0)
          flFakt = 1.0;
        uTableWidth = 0;
        for (unsigned int n=0; n<cols.size(); n++)
        {
          if (cols[n].nWidth == 0)
          {
            cols[n].nWidth = cols[n].nWidthOneLineMax * flFakt;
            if (cols[n].nWidth > nWMax)
              cols[n].nWidth = nWMax;
            if (cols[n].nWidth < cols[n].nWidthLongestTokenMax)
              cols[n].nWidth = cols[n].nWidthLongestTokenMax;
          }
          uTableWidth += cols[n].nWidth + nCellWidthMin;
        }
        uTableWidth -= nCellWidthMin;        
        if (uTextWidth < uTableWidth)
          uTextWidth = uTableWidth;
      }

      token = "";
    }
    
    fParSep = (token.compare("</p>") == 0);
    fNL = (token.compare("<br/>") == 0 || fParSep);
    if (fNL || token.compare("<p>") == 0)
    {
      uTokenW = 0;
      token = "";
    }
    else
      uTokenW = (int)(fl_width( token.c_str() ));

    if (uLineW + uTokenW > (unsigned int)nWMax || fNL || (nCellsPerRow > 0 && fLineUsed))
    {
      // Am Ende der Zeile etwas Leerraum einfügen:
      if (fNL)
        uLineW += uSpaceW;
      
      if (uLineW > uTextWidth)
        uTextWidth = uLineW;
            
      Y += (int)(labelsize());
      if (token.length())
      {
        uLineW = uTokenW + uSpaceW;
      }
      else
      {
        uLineW = 0;
      }
      X = nX0;
      fLineUsed = false; 
    }
    else
    {
      uLineW += uTokenW + uSpaceW;
    }
  
    if (token.length())
    {
      if (fPrint)
      {
        if (token.find("<ID") == 0 || token.find("<FILE") == 0)
        {
          link.x    = X;
          link.y    = Y+eRAND-y();
          link.w    = uTokenW;
          link.dest = token;
          links.push_back(link);

          fl_color(FL_RED);
          fl_draw(token.c_str(), X, link.y + y() + (int)(labelsize()) - ((int)(fl_descent())>>2));
          fl_color(textcolor);
        }
        else 
        {
          fl_draw(token.c_str(),
                  X, Y+eRAND + (int)(labelsize()) - ((int)(fl_descent())>>2));
        }
        X += uTokenW + uSpaceW;
      }
      fLineUsed = true;
    }
    else if (nCellsPerRow > 0)  // table data
    {
      int nYTop = Y;
      int nCol = 0;
      int nXCell = nX0;
      int nRowHeightMax = 0;
      for (unsigned int n=0; n<table.size(); n++)
      {
        // Layout/print this cell
        {
          unsigned int uWidthOfThisCell;
          int nThisCellsHeight = LayoutPrint(fPrint,
                                             textcolor,
                                             table[n].strText,  // text
                                             nXCell, Y,         // origin
                                             cols[nCol].nWidth, cols[nCol].nWidth,  // max width
                                             uWidthOfThisCell
                                             );
          
          // track max height of cells in this row
          if (nRowHeightMax < nThisCellsHeight)
            nRowHeightMax = nThisCellsHeight;
        }
        
        nXCell += cols[nCol].nWidth + nCellWidthMin;
        if (++nCol == nCellsPerRow)
        {
          nXCell = nX0;
          nCol = 0;
          Y += nRowHeightMax + 2*eRAND;
          if (n < table.size()-1 && fPrint)
            fl_line(nXCell, Y, nXCell+uTableWidth, Y);
          nRowHeightMax = 0;
        }
      }
      if (fPrint)
      {
        int nXVert = nX0;
        int nHalf = nCellWidthMin / 2;
        for (unsigned int n=0; n<cols.size()-1; n++)
        {
          nXVert += cols[n].nWidth;
          fl_line(nXVert+nHalf, nYTop, nXVert+nHalf, Y);
          nXVert += nCellWidthMin;
        }
      }
      fNL = false;
      nNLEnd = 0;
    }
        
    if (fParSep)
    {
      Y += (int)(labelsize());
      fLineUsed = false;
    }
    
    if (fNL)
      nNLEnd++;
    else if (token.length())
      nNLEnd = 0;
  }
  
  if (nNLEnd)
    Y -= nNLEnd * (int)(labelsize());
  if (fLineUsed)
    Y += (int)(labelsize());
  
  if (uLineW > uTextWidth)
    uTextWidth = uLineW;
  
  return(Y - nY0);
}

void TextBox::doTextLayout(bool fPrint,
                           int  nW)
{
  Fl_Color textcolor = FL_FOREGROUND_COLOR;
  
  nW -= 20; // scrollbar
  
  if (nW < 0)
    nW = w();
  
  int nWMax = nW - 2 * eRAND;
  {
    int nWM2 = 70 * (int)(fl_width("e"));
    if (nWMax > nWM2)
      nWMax = nWM2;
  }
 
  int X, Y;
  if (fPrint)
  {
    X = x();
    Y = y();

    if (fMarked)
    {
      fl_color(FL_SELECTION_COLOR);
      textcolor = fl_contrast(FL_FOREGROUND_COLOR, FL_SELECTION_COLOR);
    }
    else
    {
      fl_color(FL_BACKGROUND2_COLOR);
      textcolor = FL_FOREGROUND_COLOR;
    }
    fl_rectf(X, Y, w(), h());
    
    fl_color(textcolor);
    fl_rect(X, Y, w(), h());
    
    X += eRAND;
  }
  else
  {
    X = 0;
    Y = 0;
  }
  fl_font(labelfont(), labelsize());
  
  while (links.size())
    links.pop_back();
    
  uTextWidth = 0;
  uTextHeight = LayoutPrint(fPrint, textcolor, descr, X, Y, nW, nWMax, uTextWidth);
    
  fTextLayoutDone = true;
}

std::string TextBox::getLinkText()
{
  if (nClickedLink >= 0)
    return(links[nClickedLink].dest);
  else
    return("");
}
