// -*- mode: c; mode: fold -*-

/*
 * Copyright (C) 2008, 2011  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "texttree.h"


#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <iostream>

#include "texttreeitem.h"
#include "../shared/lib_conversions.h"
#include "../shared/utf8.h"

TextTreeBrowser::TextTreeBrowser(int X,int Y,int W,int H,const char*l, TextTreeItem* mydata)
: Fl_Browser_(X,Y,W,H,l)
{
  type(FL_SELECT_BROWSER);
  data = mydata;
}

void* TextTreeBrowser::item_first() const                    
{
  return((void*)data);
}

void* TextTreeBrowser::item_next(void *l) const
{
  if (l == 0)
    return(0);
  else
    return(((TextTreeItem*)l)->item_next());
}

void* TextTreeBrowser::item_prev(void *l) const
{
  if (l == 0)
    return(0);
  else
    return(((TextTreeItem*)l)->item_prev());
}

int   TextTreeBrowser::item_height(void *l) const            
{
  return(nItemHeight);
}

void  TextTreeBrowser::textsize(uchar s)                     
{
  Fl_Browser_::textsize(s);
  
  // calculate things  
  nItemHeight = (10*s) >> 3;
  
  // Das Quadrat sollte eine ungerade Höhe haben,
  // damit das Kreuz darin wirklich in der Mitte ist.
  nSizeBox = s;
  if ((nSizeBox & 1) == 0)
    nSizeBox--;
  
  nLengthDash = (s * 3) >> 3;  
  if ((nLengthDash & 1) == 0)
    nLengthDash--;
  nD          = (( (10*s) >> 3) - nSizeBox) >> 1;
  nTextsize   = s;
  nGroupAdd   = (s * 3) >> 1;
  
  std::cout << "TextTreeBrowser layout:\n";
  std::cout << "  nTextsize   = " << nTextsize << "\n";
  std::cout << "  nItemHeight = " << nItemHeight << "\n";
  std::cout << "  nSizeBox    = " << nSizeBox << "\n";
  std::cout << "  nLengthDash = " << nLengthDash << "\n";
  std::cout << "  nD          = " << nD << "\n";
  std::cout << "  nGroupAdd   = " << nGroupAdd << "\n";
}

int   TextTreeBrowser::handle(int e)                         
{
  TextTreeItem *l;
  
  // get box INSIDE the scrollbars
  int X,Y,W,H;  
  bbox(X,Y,W,H);
  
  switch (e) 
  {
   case FL_PUSH:
    // to make the Fl_Browser_ handle the scrollbars:
    if (
        !Fl::event_inside(X,Y,W,H)
        )
      break;
    
    return 1;
    break;
    
   // only to make sure Fl_Browser_ will not handle it
   case FL_DRAG:    
    return(1);    
    
   case FL_RELEASE:
    l = (TextTreeItem*)find_item(Fl::event_y());
    if (l != (TextTreeItem*)0)
    {
      int  nXRelease = Fl::event_x() + hposition() - leftedge();
      bool fRedraw   = false;
      int  nXBoxMin  = 3 + l->treeDepth() * 2 * nTextsize;
      int  nXBoxMax  = nXBoxMin + nTextsize;
      
      if (l->itemFirstChild != 0 &&
          l->itemParent     != 0 &&
          nXRelease >= nXBoxMin)
      {
        if (nXRelease < nXBoxMax)
          l->fExpanded = !l->fExpanded;
        fRedraw = true;
      }        
      
      if (nXRelease >= nXBoxMax)
      {
        selected_path = l->getPath();
        do_callback(this, user_data());
      }
        
      if (fRedraw)
        redraw();
      
      return(1);        
    }
  }
  return Fl_Browser_::handle(e);
}

int   TextTreeBrowser::item_width(void *v) const             
{
  TextTreeItem* item = (TextTreeItem*)v;
  int           X    = 3 + item->treeDepth() * 2 * nTextsize;

  if (item->itemFirstChild != 0)
    X += nGroupAdd;
    
  return( (int)(fl_width( item->getText().c_str() )) + X);
}

void  TextTreeBrowser::item_draw(void *v, 
                                int X, int Y, 
                                int, int) const             
{
  TextTreeItem* item = (TextTreeItem*)v;
  Fl_Color      lcol = textcolor();
  
  fl_color(lcol);
  if (v == selection())
  {
    fl_color(fl_contrast(lcol, FL_SELECTION_COLOR));
  }
    
  X += 3;
  
  X += item->treeDepth() * 2 * nTextsize;
  
  if (item->itemFirstChild != 0)
  {
    int  nMitte   = (nSizeBox>>1);
    int  nAbstand = (nSizeBox - nLengthDash)>>1;        
    
    fl_rect(X, Y+nD, nSizeBox, nSizeBox);
    
    if (item->fExpanded)
      fl_line(X+nAbstand, Y+nD+nMitte,   X+nAbstand+nLengthDash-1, Y+nD+nMitte);
    else
    {
      fl_xyline(X+nAbstand, Y+nD+nMitte,   X+nAbstand+nLengthDash-1);
      fl_yxline(X+nMitte,   Y+nD+nAbstand, Y+nD+nAbstand+nLengthDash-1);
    }
    X += (nTextsize * 3) >> 1;
  }
  
  fl_font(textfont(), nTextsize);
  fl_draw(item->getText().c_str(),
          X, Y + nTextsize - ((int)(fl_descent())>>2));
  
  // std::cout << "draw " << X << ", " << Y << ": " << item->getText() << "\n";
}

void TextTreeBrowser::setData(TextTreeItem* newdata)
{
  data = newdata;
  
  // Tell Fl_Browser_ about it:
  new_list();
  redraw_lines();
}

void TextTreeBrowser::delData()
{
  if (data != 0)
    delete data; 
}

void TextTreeBrowser::show(std::string text)
{
  TextTreeItem* match = data->find(UTF8::to_lower(text));
  if (match)
  {
    TextTreeItem* parent = match->itemParent;
    while (parent)
    {
      parent->fExpanded = true;
      parent = parent->itemParent;
    }
    // std::cout << "result is " << match->getPath() << "\n";
    select_only((void*)match);
    // how many items are shown?
    int nNumLinesHalf = h() / (item_height(match)*3);
    TextTreeItem* p = (TextTreeItem*)(match->item_next());
    while (nNumLinesHalf && p)
    {
      match = p;
      p = (TextTreeItem*)(match->item_next());
      nNumLinesHalf--;
    }
    display((void*)match);
    redraw();
  }
}

TextTreeItem* TextTreeBrowser::getSelectedItem()
{
  return((TextTreeItem*)selection());
}

TextTreeBrowser::~TextTreeBrowser()
{
  delData(); 
}
