/*
 * Copyright (C) 2003-2007, 2011  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "SimpleXMLTransfer.h"
#include "lib_conversions.h"
#include <ctype.h>
#include <iostream>
#include <stdlib.h>
#include <cstring>

#ifdef WIN32_FLTK13
# include <FL/fl_utf8.h>
# include "filestream.h"
#endif

// 1: ?
// 2: show compare
#define DEBUG 0

#if DEBUG != 0
# include <stdio.h>
#endif


/**
 * Only returns true for characters below 0x80 and therefore never
 * for multibyte UTF-8 characters.
 */
inline bool isspace_utf8(unsigned char ucChar)
{
  return( (ucChar & 128) == 0 && isspace(ucChar) != 0 );
}

const char* XMLException::what()
{
  return (myMessage.c_str());
}

XMLException::XMLException(std::string message)
{
  myMessage = message;
}




SimpleXMLTransfer::SimpleXMLTransfer()
{
  myName      = "data";
  content     = (std::string*) 0;
#if DEBUG != 0
  sourcedescr = "default constructor";
#endif
}

SimpleXMLTransfer::SimpleXMLTransfer(SimpleXMLTransfer* source)
{
  content     = (std::string*) 0;
#if DEBUG != 0
  sourcedescr = "copy constructor";
#endif
  
  setName(source->getName());
  
  for (int n=0; n<source->getAttributeCount(); n++)
    addAttribute(source->attributeName(n), source->attributeVal(n));
  
  for (int n=0; n<source->getChildCount(); n++)
    addChild(new SimpleXMLTransfer(source->getChildAt(n)));
  
  if (source->getContentString().length())
  {
    content = new std::string();
    *content = source->getContentString();
  }
}

SimpleXMLTransfer::SimpleXMLTransfer(std::istream& in)
{
  content = (std::string*) 0;
#if DEBUG != 0
  sourcedescr = "stream";
#endif
  
  char* acBuf;
  int   nBufSize = 2048;
  int   nBufFill = 0;
  int   nBufIdx  = 0;
  acBuf = (char*)malloc(nBufSize);        
  readStream(in, acBuf, nBufSize, nBufFill, nBufIdx);
  free(acBuf);
}

SimpleXMLTransfer::SimpleXMLTransfer(std::istream& in, std::string strDescr, char*& acBuf, int& nBufSize, int& nBufFill, int& nBufIdx)
{
  content     = (std::string*) 0;
#if DEBUG != 0
  sourcedescr = strDescr;
#endif
  
  readStream(in, acBuf, nBufSize, nBufFill, nBufIdx);
}

SimpleXMLTransfer::SimpleXMLTransfer(std::string source)
{
  content = (std::string*) 0;
    
#ifdef WIN32_FLTK13
  FILE* fp = fl_fopen(source.c_str(), "rb");
  if (fp == 0)
    throw XMLException("Error opening " + source);
  iFILEstream in(fp);
#else
  std::ifstream in;  
  in.open(source.c_str());
#endif
  
#if DEBUG != 0
  sourcedescr = "file: " + source;
#endif
  
  if (!in)
  {
    std::cerr << "Error opening " << source << "\n";
    throw XMLException("Error opening " + source);
  }
  else
  {
    char* acBuf;
    int   nBufSize = 2048;
    int   nBufFill = 0;
    int   nBufIdx  = 0;
    acBuf = (char*)malloc(nBufSize);        
    if (acBuf == 0)
      throw XMLException("Error allocating memory");    
    in.clear();
    readStream(in, acBuf, nBufSize, nBufFill, nBufIdx);
    free(acBuf);
        
#ifdef WIN32_FLTK13
    // std::cout << "close\n"; std::cout.flush();
#else
    in.close();
#endif
  }
}

SimpleXMLTransfer::~SimpleXMLTransfer()
{
  SimpleXMLTransfer* tmp;
  
  if (content != (std::string*) 0)
    delete content;
  
  while (children.size() > 0)
  {
    tmp = children[children.size() - 1];
    delete tmp;
    children.pop_back();
  }
}

void SimpleXMLTransfer::addAttribute(std::string const& attributeName,
                                     long int    attributeVal)
{
  addAttribute(attributeName, itoStr(attributeVal, ' ', 1));
}

void SimpleXMLTransfer::addAttribute(std::string const& attributeName,
                                     std::string const& attributeVal)
{
  //       System.out.println("SimpleXMLTransfer.addAttribute(" + attributeName + ", " + attributeVal + ")");
  //
  attrName.push_back(attributeName);
  attrVal.push_back(attributeVal);
}

void SimpleXMLTransfer::addChild(SimpleXMLTransfer*  child)
{
  children.push_back(child);
}

std::string SimpleXMLTransfer::attribute(std::string attr)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
  {
    throw XMLException("No Attribute named " + attr + " in " + getName());
  }
  else
  {
    return (attrVal[index]);
  }
}

std::string SimpleXMLTransfer::attribute(std::string attr,
					 std::string defaultVal)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
    return (defaultVal);
  else
    return (attrVal[index]);
}

double SimpleXMLTransfer::attributeAsDouble(std::string attr)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
  {
    throw XMLException("No Attribute named " + attr + " in " + getName());
  }
  else
  {
    return (convToDouble(attrVal[index]));
  }
}

double SimpleXMLTransfer::attributeAsDouble(std::string attr, 
                                            double      defaultVal)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
  {
    return (defaultVal);
  }
  else
  {
    return (convToDouble(attrVal[index]));
  }
}

double SimpleXMLTransfer::attributeAsDouble(unsigned int index)
{
  if (index > attrName.size())
  {
    throw XMLException("No such attribute in " + getName());
  }
  else
  {
    return (convToDouble(attrVal[index]));
  }
}

float SimpleXMLTransfer::attributeAsFloat(std::string attr)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
  {
    throw XMLException("No Attribute named " + attr + " in " + getName());
  }
  else
  {
    return (convToFloat(attrVal[index]));
  }
}

int SimpleXMLTransfer::attributeAsInt(std::string attr)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
  {
    throw XMLException("No Attribute named " + attr + " in " + getName());
  }
  else
  {
    return (convToInt(attrVal[index]));
  }
}

int SimpleXMLTransfer::attributeAsInt(std::string attr, int defaultVal)
{
  int index = indexOfAttribute(attr);
  
  if (index < 0)
  {
    return (defaultVal);
  }
  else
  {
    return (convToInt(attrVal[index]));
  }
}

std::string SimpleXMLTransfer::attributeName(unsigned int index)
{
  if (index > attrName.size())
    throw XMLException("No such attribute in " + getName());
  else
    return(attrName[index]);
}

std::string SimpleXMLTransfer::attributeVal(unsigned int index)
{
  if (index > attrVal.size())
    throw XMLException("No such attribute in " + getName());
  else
    return(attrVal[index]);
}

void SimpleXMLTransfer::convFromXML(char* pBuf, int nLen)
{
  char ch;
  char* pOut = pBuf;

  while (nLen>0)
  {
    ch = *pBuf;
    
    if (ch != '&' 
      /*  || (ch == '&' && in.find(';', i) == std::string::npos) // enables reading faulty xml?  */
        )
    {
      *pOut = *pBuf;
      pOut++;
      pBuf++;
      nLen--;
    }
    else
    {
      pBuf++;
      nLen--;
      
      ch = *pBuf;
        
      if (ch == 'l')       // lt
        *pOut = '<';
      else if (ch == 'g')  // gt
        *pOut = '>';
      else if (ch == 'a')  // amp
      {
        pBuf++;
        nLen--;
        ch = *pBuf;
        if (ch == 'm')
          *pOut = '&';
        else
          *pOut = '\'';
      }
      else if (ch == 'q')  // quot
        *pOut = '\"';
      else if (ch != '#')
      {
        std::string strErr;
        strErr.assign(pBuf, 3);
        throw XMLException("Error converting character: " + strErr);
      }
      else
      {
        pBuf+= 2;
        int val = hex2int(*pBuf);
        pBuf++;
        nLen-= 3;
        
        while (*pBuf != ';')
        {
          val = (val<<4) + hex2int(*pBuf);
          pBuf++;
          nLen--;
        }                
        *pOut = (char)(val);
      }
      pOut++;
      while (*pBuf != ';')
      {
        pBuf++;
        nLen--;
      }
      pBuf++;
      nLen--;
    }
  }
  *pOut = '\0';
}

double SimpleXMLTransfer::convToDouble(std::string value)
{
  char*       ptr;
  double      tmp;
  const char* inptr = value.c_str();
  
  tmp = strtod(inptr, &ptr);
  if (ptr == inptr || *ptr != '\0')
  {
    value = "convToDouble: no number or trailing garbage in " + value;
    throw XMLException(value.c_str());
  }
  return(tmp);
}

float SimpleXMLTransfer::convToFloat(std::string value)
{
  return(convToDouble(value));
}

int SimpleXMLTransfer::convToInt(std::string value)
{
  long        nVal;
  char*       ptr;
  const char* inptr;
  int         base;
  
  if (value.find("0b") == 0 && value.length() > 2)
  {
    // binäre Darstellung
    inptr = value.c_str()+2;
    base  = 2;
  }
  else if (value.length() > 1 && toupper(value[1]) == 'X')
  {
    // hexadezimale Darstellung
    inptr = value.c_str()+2;
    base  = 16;
  }
  else
  {
    // dezimale Darstellung
    inptr = value.c_str();
    base  = 10;
  }
  
  nVal = strtol(inptr, &ptr, base);
  if (ptr == inptr || *ptr != '\0')
  {
    value = "convToInt: no number or trailing garbage in " + value;
    throw XMLException(value.c_str());
  }
  return(nVal);
}

std::string* SimpleXMLTransfer::convToXML(std::string* in, std::string* strBuf)
{
  int length = in->length();
  char ch;
  strBuf->clear();
  
  for (int i = 0; i < length; i++)
  {
    ch = in->at(i);

    // Es wird nur das absolut nötigste konvertiert. Ich gehe ja von UTF-8 aus, da sollte 
    // der komplette Rest erlaubt sein.
    switch (ch)
    {
      case '<':
        if (strBuf->empty())
          *strBuf = in->substr(0, i);
        strBuf->append("&lt;");
        break;
      
      case '>':
        if (strBuf->empty())
          *strBuf = in->substr(0, i);
        strBuf->append("&gt;");
        break;

      case '&':
        if (strBuf->empty())
          *strBuf = in->substr(0, i);
        strBuf->append("&amp;");
        break;
        
      case '\'':
        if (strBuf->empty())
          *strBuf = in->substr(0, i);
        strBuf->append("&apos;");
        break;
        
      case '"':
        if (strBuf->empty())
          *strBuf = in->substr(0, i);
        strBuf->append("&quot;");
        break;
        
      case '\n':
      case '\r':
      case '\t':
        if (strBuf->empty())
          *strBuf = in->substr(0, i);

        strBuf->append("&#x");
        strBuf->append(itoHexStr( (int)((unsigned char)(ch)) ));
      
        strBuf->push_back(';');
        break;

      default:
        if (!strBuf->empty())
          strBuf->push_back(ch);
    }
  }
  if (strBuf->empty())
    return(in);
  else
    return(strBuf);
}

void SimpleXMLTransfer::delAttribute(std::string attribute)
{
  int n = indexOfAttribute(attribute);
  if (n < 0)
    throw XMLException("Attribute named " + attribute +
		       " does not exist in " + getName());
  
  attrName.erase(attrName.begin() + n);
  attrVal.erase(attrVal.begin() + n);
}

bool SimpleXMLTransfer::equals(SimpleXMLTransfer* item)
{
  int                nSize;
  int                nSizeB;
  SimpleXMLTransfer* itA;
  SimpleXMLTransfer* itB;
  bool               fFound;
  
  // --- Compare name -------------------------------------
  if (getName().compare(item->getName()))
  {
#if DEBUG == 2
    printf("Name differs: %s != %s\n", getName().c_str(), item->getName().c_str());
#endif
    return(false);
  }
  
  // --- Compare attributes -------------------------------
  // Number of attributes
  nSize = getAttributeCount();
  if (nSize != item->getAttributeCount())
  {
#if DEBUG == 2
    printf("different number of attributes\n");
#endif
    return(false);
  }
  // If we have the same number of attributes, it is OK
  // if my attributes can be found and are equal to those 
  // of item.  
  for (int n=0; n<nSize; n++)
  {
    if (item->indexOfAttribute(attrName[n]) == -1)
    {
#if DEBUG == 2
      printf("No attribute %s\n", attrName[n].c_str());
#endif
      return(false);
    }
    
    if (item->attribute(attrName[n]).compare(attrVal[n]))
    {
#if DEBUG == 2
      printf("%s: %s != %s\n", attrName[n].c_str(), attrVal[n].c_str(), item->attribute(attrName[n]).c_str());
#endif
      return(false);
    }
  }
  
  // --- Compare children ---------------------------------
  // Number of children
  nSize = getChildCount();
  if (nSize != item->getChildCount())
  {
#if DEBUG == 2
    printf("different number of children\n");
#endif
    return(false);
  }
  
  // name and content
  nSizeB = item->getChildCount();
  for (int n=0; n<nSize; n++)
  {
    itA = children[n];
    // As order does not matter, I have to compare to every child
    // of 'item'.
    fFound = false;
    for (int m=0; m<nSizeB && !fFound; m++)
    {
      itB = item->getChildAt(m);
      if (itA->equals(itB))
        fFound = true;
    }
    if (!fFound)
      return(false);
  }
  
  return(true);
}

bool SimpleXMLTransfer::equalsOrdered(SimpleXMLTransfer* item)
{
  int                nSize;
  SimpleXMLTransfer* itA;
  SimpleXMLTransfer* itB;
  
  // --- Compare name -------------------------------------
  if (getName().compare(item->getName()))
  {
#if DEBUG == 2
    printf("Name differs: %s != %s\n", getName().c_str(), item->getName().c_str());
#endif
    return(false);
  }
  
  // --- Compare attributes -------------------------------
  // Number of attributes
  nSize = getAttributeCount();
  if (nSize != item->getAttributeCount())
  {
#if DEBUG == 2
    printf("different number of attributes\n");
#endif
    return(false);
  }
  // If we have the same number of attributes, it is OK
  // if my attributes can be found and are equal to those 
  // of item.
  for (int n=0; n<nSize; n++)
  {
    if (item->indexOfAttribute(attrName[n]) == -1)
    {
#if DEBUG == 2
      printf("No attribute %s\n", attrName[n].c_str());
#endif
      return(false);
    }
    
    if (item->attribute(attrName[n]).compare(attrVal[n]))
    {
#if DEBUG == 2
      printf("%s: %s != %s\n", attrName[n].c_str(), attrVal[n].c_str(), item->attribute(attrName[n]).c_str());
#endif
      return(false);
    }
  }
  
  // --- Compare children ---------------------------------
  // Number of children
  nSize = getChildCount();
  if (nSize != item->getChildCount())
  {
#if DEBUG == 2
    printf("different number of children\n");
#endif
    return(false);
  }
  
  // name and content
  for (int n=0; n<nSize; n++)
  {
    itA = children[n];
    itB = item->getChildAt(n);
    if (itA->equalsOrdered(itB) == false)
      return(false);
  }
  
  return(true);
}

SimpleXMLTransfer* SimpleXMLTransfer::getChild(std::string child, bool fCreate)
{
  int                    index;
  SimpleXMLTransfer*     tmp;
  std::string::size_type pos;
  
  pos = child.find('.');
  
  if (pos != std::string::npos)
  {
    std::string parentName = "";
    std::string childName = "";
    SimpleXMLTransfer* item = this;
    
    do
    {
      childName = child.substr(0, pos);
      child     = child.substr(pos + 1);
      pos       = child.find('.');
      
      index = item->indexOfChild(childName);
      if (index < 0)
      {
        if (fCreate)
        {
          tmp = new SimpleXMLTransfer();
          tmp->setName(childName);
          item->addChild(tmp);
          index = item->indexOfChild(childName);
        }
        else
          throw XMLException("No item named " + childName + " in " +
                             parentName);
      }
      
      if (parentName.length() > 0)
        parentName.push_back('.');
      parentName.append(childName);
      
      item = item->getChildAt(index);
    }
    while (pos != std::string::npos);
    
    index = item->indexOfChild(child);
    if (index < 0)
    {
      if (fCreate)
      {
        tmp = new SimpleXMLTransfer();
        tmp->setName(child);
        item->addChild(tmp);
        return(tmp);
      }
      else
        throw XMLException("No item named " + child + " in " + parentName);
    }
    
    return (item->getChildAt(index));
  }
  else
  {
    index = indexOfChild(child);
    
    if (index < 0)
    {
      if (fCreate)
      {
        tmp = new SimpleXMLTransfer();
        tmp->setName(child);
        addChild(tmp);
        return(tmp);
      }
      else
        throw XMLException("No child named " + child + " in " + getName());
    }
    else
    {
      return (children[index]);
    }
  }
}

SimpleXMLTransfer* SimpleXMLTransfer::getChild(std::string child, int& idx)
{
  if (idx >= getChildCount()-1)
  {
    return(0);
  }
  else
  {
    SimpleXMLTransfer* tmp;
    
    for (int i=idx+1; i<getChildCount(); i++)
    {
      tmp = getChildAt(i);
      if (tmp->getName().compare(child) == 0)
      {
        idx = i;
        return(tmp);
      }
    }
    return(0);
  }
}

SimpleXMLTransfer* SimpleXMLTransfer::getChildAt(int childIndex)
{
  if (children.size() == 0)
    return ((SimpleXMLTransfer *) 0);
  return (children[childIndex]);
}

int SimpleXMLTransfer::getChildCount() const
{
  return (children.size());
}

double SimpleXMLTransfer::getContentDouble()
{
  return (convToDouble(*content));
}

int SimpleXMLTransfer::getContentInt()
{
  return (convToInt(*content));
}

std::string SimpleXMLTransfer::getContentString()
{
  if (content != (std::string*) 0)
    return (*content);
  else
    return ("");
}

double SimpleXMLTransfer::getDouble(std::string path)
{
  std::string::size_type pos = path.rfind('.');
  
  if (pos != std::string::npos)
  {
    SimpleXMLTransfer* item = getChild(path.substr(0, pos));
    return (item->attributeAsDouble(path.substr(pos + 1)));
  }
  else
  {
    return (attributeAsDouble(path));
  }
}

double SimpleXMLTransfer::getDouble(std::string path, double dDefault)
{
  try
  {
    std::string::size_type pos = path.rfind('.');
    
    if (pos != std::string::npos)
    {
      SimpleXMLTransfer* item = getChild(path.substr(0, pos));
      return (item->attributeAsDouble(path.substr(pos + 1)));
    }
    else
    {
      return (attributeAsDouble(path));
    }
  }
  catch (XMLException e)
  {
    makeSureAttributeExists(path, doubleToString(dDefault).c_str());
    return(dDefault);
  }
}

int SimpleXMLTransfer::getInt(std::string path)
{
  std::string::size_type pos = path.rfind('.');
  
  if (pos != std::string::npos)
  {
    SimpleXMLTransfer* item = getChild(path.substr(0, pos));
    return (item->attributeAsInt(path.substr(pos + 1)));
  }
  else
  {
    return (attributeAsInt(path));
  }
}

int SimpleXMLTransfer::getInt(std::string path, int nDefault)
{
  try
  {
    std::string::size_type pos = path.rfind('.');
    
    if (pos != std::string::npos)
    {
      SimpleXMLTransfer* item = getChild(path.substr(0, pos));
      return (item->attributeAsInt(path.substr(pos + 1)));
    }
    else
    {
      return (attributeAsInt(path));
    }
  }
  catch (XMLException e)
  {
    makeSureAttributeExists(path, itoStr(nDefault, ' ', 1).c_str());
    return(nDefault);
  }
}

std::string SimpleXMLTransfer::getName()
{
  return (myName);
}

std::string SimpleXMLTransfer::getString(std::string path)
{
  std::string::size_type pos = path.rfind('.');
  
  if (pos != std::string::npos)
  {
    SimpleXMLTransfer* item = getChild(path.substr(0, pos));
    return (item->attribute(path.substr(pos + 1)));
  }
  else
  {
    return (attribute(path));
  }
}

std::string SimpleXMLTransfer::getString(std::string path, std::string stringDefault)
{
  try
  {
    std::string::size_type pos = path.rfind('.');
    
    if (pos != std::string::npos)
    {
      SimpleXMLTransfer* item = getChild(path.substr(0, pos));
      return (item->attribute(path.substr(pos + 1)));
    }
    else
    {
      return (attribute(path));
    }
  }
  catch (XMLException e)
  {
    makeSureAttributeExists(path, stringDefault.c_str());
    return(stringDefault);
  }
}

int SimpleXMLTransfer::indexOfAttribute(std::string attr) const
{
#if DEBUG == 1
  printf("int SimpleXMLTransfer::indexOfAttribute(\"%s\")\n", attr.c_str());
#endif
  if (attrName.size() == 0)
    return (-1);
  
  int index = -1;
  
  for (unsigned int i = 0; i < attrName.size() && index == -1; i++)
    if (attr.compare(attrName[i]) == 0)
      index = i;
  
#if DEBUG == 1
  printf("int SimpleXMLTransfer::indexOfAttribute(\"%s\") = %i\n",
	 attr.c_str(), index);
#endif
  return (index);
}

int SimpleXMLTransfer::indexOfChild(std::string child)
{
  if (children.size() == 0)
    return (-1);
  
  int index = -1;
  int size  = children.size();
  
  for (int i = 0; i < size && index == -1; i++)
    if (child.compare(children[i]->getName()) == 0)
      index = i;
  
  return (index);
}

int SimpleXMLTransfer::indexOfChild(std::string child, int nStartIdx)
{
  if (children.size() == 0)
    return (-1);
  
  if (nStartIdx < 0 || nStartIdx >= (int)(children.size()))
    nStartIdx = 0;
  
  int index = -1;
  int size  = children.size();
  
  for (int i = nStartIdx; i < size && index == -1; i++)
    if (child.compare(children[i]->getName()) == 0)
      index = i;
  
  if (index < 0)
  {
    for (int i = 0; i < nStartIdx && index == -1; i++)
      if (child.compare(children[i]->getName()) == 0)
        index = i;
  }
  
  return (index);  
}

void SimpleXMLTransfer::makeSureAttributeExists(std::string attribute, const char* defaultVal)
{
  if (attribute.find('.') != std::string::npos)
  {
    int index;
    std::string childName;
    SimpleXMLTransfer* item = this;
    std::string::size_type pos;
    
    while ((pos = attribute.find('.')) != std::string::npos)
    {
      childName = attribute.substr(0, pos);
      attribute = attribute.substr(pos + 1);
      
      index = item->indexOfChild(childName);
      if (index < 0)
      {
        SimpleXMLTransfer* tmp = new SimpleXMLTransfer();
        tmp->setName(childName);
        item->addChild(tmp);
        item = tmp;
      }
      else
        item = item->getChildAt(index);
    }
    
    item->makeSureAttributeExists(attribute, defaultVal);
  }
  else
  {
    int index = indexOfAttribute(attribute);
    if (index < 0)
    {
      addAttribute(attribute, defaultVal);
    }
  }
}

void SimpleXMLTransfer::merge(SimpleXMLTransfer* b)
{
  int                idx;
  SimpleXMLTransfer* bchild;
  
  // Die Attribute von b bei mir einbauen
  for (int n=0; n<b->getAttributeCount(); n++)
  {
    idx = indexOfAttribute(b->attributeName(n));
    if (idx != -1)
    {
      throw XMLException("doppeltes Attribut " + b->attributeName(n) + " in " + getName());
    }
    addAttribute(b->attributeName(n), b->attributeVal(n));
  }
  
  // Alle Kinder von b durchgehen. Wenn ein Kind mit diesem Namen auch bei mir vorkommt, 
  // füge ich die beiden zusammen. Wenn das Kind von b einen neuen Namen hat, wird es 
  // einfach hinzugefügt.
  for (int n=0; n<b->getChildCount(); n++)
  {
    bchild = b->getChildAt(n);
    idx    = indexOfChild(bchild->getName());
    if (idx == -1)
      addChild(bchild);
    else
      getChildAt(idx)->merge(bchild);
  }
}

void SimpleXMLTransfer::overwriteAttribute(unsigned int index, std::string val)
{
  if (index > attrName.size())
    throw XMLException("No such attribute in " + getName());
  else
    attrVal[index] = val;
}

void SimpleXMLTransfer::print(std::ostream& out, int nIndent, int nIndentStep, int nLineBreak)
{
  out << "<?xml version=\"1.0\" ?>\n";
  
  printNoHeader(out, nIndent, nIndentStep, nLineBreak);
}

void SimpleXMLTransfer::print(std::ostream& out,
                              int           nIndent,
                              std::string   val)
{
  for (int n = 0; n < nIndent; n++)
    out << " ";

  out << val;  
}

void SimpleXMLTransfer::printNoHeader(std::ostream& out, int nIndent, int nIndentStep, int nLineBreak)
{
  int nCMax = -1;
  
  if (comment.size() > 0 && commentPos[0] == -1)
  {
    print(out, nIndent*nIndentStep, comment[0] + "\n");
    nCMax = 0;
  }
  
  // Es könnte sein, dass dieses Element ein reiner Kommentar ist!
  if (getName().length() > 0)
  {
    // mich selbst eröffnen
    print(out, nIndent*nIndentStep, "<" + getName());
    
    // Attribute schreiben
    if (attrName.size() > 0)
    {
      int nSpalte = nIndent*nIndentStep + getName().length();
      std::string* pVal;
      std::string aVal;
      
      for (unsigned int i = 0; i < attrName.size(); i++)
      {
        pVal = convToXML(&(attrVal[i]), &aVal);
        std::string& aName = attrName[i];
        
        if (nLineBreak > 0 && (int)(nSpalte + pVal->length() + aName.length()) > nLineBreak)
        {
          out << "\n";
          
          for (int n = 0; n < nIndent*nIndentStep; n++)
            out << " ";
          
          out << " ";
          nSpalte = (nIndent + 1)*nIndentStep;
        }
        else
          nSpalte += pVal->length() + aName.length() + 4;
        
        out << " " << aName << "=\"" << (*pVal) << "\"";
      }
    }
    
    if (content != (std::string *) 0 || children.size() > 0)
    {
      out << ">";
      
      // Wenn ich irgendwelchen Inhalt habe, schreibe ich den vor den Kindern
      // und den Kommentaren.
      if (content != (std::string *) 0)
      {
        out << getContentString();
      }
      
      if (comment.size() > (unsigned int)(nCMax+1))
      {
        if (commentPos[nCMax+1] == 0)
        {
          nCMax++;
          out << "\n";
          print(out, nIndent*nIndentStep, comment[nCMax]);
        }
      }
      
      // Die Kinder sollen ihre Daten schreiben, falls es welche gibt.
      if (children.size() > 0)
      {
        out << "\n";
        
        for (unsigned int i = 0; i < children.size(); i++)
        {
          children[i]->printNoHeader(out, nIndent + 1, nIndentStep, nLineBreak);
          for (unsigned int m=0; m<comment.size(); m++)
          {
            if (commentPos[m] == (int)i+1)
              print(out, nIndent*nIndentStep, comment[m] + "\n");              
          }
        }
        
        // ich bin fertig
        for (int n = 0; n < nIndent*nIndentStep; n++)
          out << " ";
      }
      out << "</" << getName() << ">\n";
    }
    else
      // es gibt keine Kinder
      out << "/>\n";
  }
}

void SimpleXMLTransfer::readStream(std::istream& in, char*& acBuf, int& nBufSize, int& nBufFill, int& nBufIdx)
{
  bool fMachen = true;
  
  std::string strTemp, strAttributeVal;
  int c0;
  T_State nState;
  T_State nStateAfterComment = STATE_IDLE;
  
  int nBufIdxStart = nBufIdx;
  
  if (nBufFill == 0)
  {
    in.read(acBuf, nBufSize);
    nBufFill = in.gcount();
    
    if (in.eof() && nBufFill < 4)
      throw XMLException("Error reading xml stream: eof");
    if (in.fail() && nBufFill < 4)
      throw XMLException("Error reading xml stream");
    
    nState = STATE_IDLE;
    if (nBufIdx >= nBufFill || nBufFill < 4) // a minimal file has to contain four characters: <X/>
    {
      throw XMLException("Internal error C: " + itoStr(nBufIdx, ' ', 1) + " >= " + itoStr(nBufFill, ' ', 1));
    }
  }
  else
  {
    nState = STATE_READNAME;
  }
  
  do
  {
    c0 = acBuf[nBufIdx];
    
    switch (nState)
    {
        // Auf eine öffnende Klammer warten.
      case STATE_IDLE:
        if (c0 == '<')
        {
          nState = STATE_START;
          nBufIdxStart = nBufIdx; // wichtig, falls ein Kommentar kommt
        }
        break;
        
        //
      case STATE_START:
        if (c0 == '?')
          nState = STATE_INITTAG;
        else if (c0 == '!')
        {
          nStateAfterComment = STATE_IDLE;
          nState = STATE_COMMENT;
        }
        else
        {
          nState = STATE_READNAME;
          nBufIdxStart = nBufIdx;
        }
        break;
        
        // Das Ende vom Inittag abwarten.
      case STATE_INITTAG:
        if (c0 == '>')
        {
          nState = STATE_IDLE;
          nBufIdxStart = nBufIdx;
        }
        break;
        
        // Das Ende vom Kommentar abwarten.
      case STATE_COMMENT:
        if (c0 == '-')
          nState = STATE_COMMENT_END1;
        break;
        
      case STATE_COMMENT_END1:
        if (c0 == '-')
          nState = STATE_COMMENT_END2;
        else
          nState = STATE_COMMENT;
        break;
        
      case STATE_COMMENT_END2:
        if (c0 == '>')
        {
          strTemp.assign(&acBuf[nBufIdxStart], 1+nBufIdx-nBufIdxStart);
          comment.push_back(strTemp);
          nBufIdxStart = nBufIdx;
          
          if (nStateAfterComment == STATE_IDLE)
            commentPos.push_back(-1);
          else
            commentPos.push_back(children.size());
          nState = nStateAfterComment;
        }
        else
          nState = STATE_COMMENT;
        break;
        
        // Den Namen einlesen
      case STATE_READNAME:
        if (c0 == '=')
        {
          throw XMLException("Found \'=\' without preceding attributename in " + myName);
        }
        else if ((c0 == '>' || c0 == '/' || c0 == ' ' || c0 == '\n' || c0 == '\r' || c0 == '\t') && nBufIdx > nBufIdxStart)
        {
          myName.assign(&acBuf[nBufIdxStart], nBufIdx-nBufIdxStart);
          nBufIdxStart = nBufIdx;
          if (c0 == '>')
          {
            nState = STATE_WAIT_CONT;
          }
          else if (c0 == '/')
            nState = STATE_EXPECT_CLOSE;
          else
            nState = STATE_WAIT_ATTR_NAME;
        }
        break;
        
      case STATE_WAIT_ATTR_NAME:
        nBufIdxStart = nBufIdx;
        if (c0 == '/')
        {
          nState = STATE_EXPECT_CLOSE;
        }
        else if (c0 == '>')
        {
          nState = STATE_WAIT_CONT;
        }
        else if (isspace(c0) == false)
          nState = STATE_ATTR_NAME;
        break;
        
      case STATE_ATTR_NAME:
        if (c0 == '=' || isspace(c0))
        {
          strTemp.assign(&acBuf[nBufIdxStart], nBufIdx-nBufIdxStart);
          if (c0 == '=')
            nState = STATE_ATTR_VAL_START;
          else
            nState = STATE_ATTR_VAL_WAIT_EQ;
          nBufIdxStart = nBufIdx;
        }
        break;
        
      case STATE_ATTR_VAL_WAIT_EQ:
        if (c0 == '=')
          nState = STATE_ATTR_VAL_START;
        else if (isspace(c0) == false)
        {
          throw XMLException(sourcedescr + ": Element " + myName +
                             ": expected \'=\', got \'" + (char)(c0) +
                             "\' after attribute " + strTemp);
        }
        break;
        
      case STATE_ATTR_VAL_START:
        nBufIdxStart = nBufIdx;
        if (c0 == '"')
          nState = STATE_ATTR_VAL1;
        break;
        
      case STATE_ATTR_VAL1:
        nBufIdxStart = nBufIdx;
        if (c0 == '"')
        {
          addAttribute(strTemp, "");
          nState = STATE_WAIT_ATTR_NAME;
        }
        else          
          nState = STATE_ATTR_VAL;
        break;
        
      case STATE_ATTR_VAL:
        if (c0 == '"')
        {
          convFromXML(&acBuf[nBufIdxStart], nBufIdx-nBufIdxStart);
          addAttribute(strTemp, &acBuf[nBufIdxStart]);
          nBufIdxStart = nBufIdx;
          nState = STATE_WAIT_ATTR_NAME;
        }
        break;
        
      case STATE_EXPECT_CLOSE:
        if (c0 == '>')
        {
          nBufIdxStart = nBufIdx;
          fMachen = false;
        }
        break;
                
      case STATE_WAIT_CONT:
        nBufIdxStart = nBufIdx;
        if (c0 == '<')
        {
          nState = STATE_CONT_START;
        }
        else
        {
          // Der Inhalt könnte von Tags unterbrochen sein, daher füge ich dessen Zeichen aneinander.
          // Das ist zwar langsamer als die ansonsten benutzte Methode, macht aber nix -- diese Form
          // von Inhalt verwende ich üblicherweise nicht.
          if (content == (std::string *) 0 && isspace_utf8(c0) == false)
            content = new std::string();
          if (content != (std::string *) 0)
            content->push_back(c0);
        }
        break;
        
      case STATE_CONT_START:
        if (c0 == '/')
        {
          nBufIdxStart = nBufIdx;
          nState = STATE_ENDTAG;
        }
        else if (c0 == '!')
        {
          nStateAfterComment = STATE_WAIT_CONT;
          nState = STATE_COMMENT;
        }
        else
        {
          SimpleXMLTransfer* child = new SimpleXMLTransfer(in, sourcedescr, acBuf, nBufSize, nBufFill, nBufIdx);
          addChild(child);
          nState = STATE_WAIT_CONT;
          if (nBufIdx == nBufFill)
          {
            fMachen = false;
          }
        }
        break;
        
        // Ich warte auf das Ende und überprüfe, ob auch
        // mein Name wieder korrekt im End-Tag steht.
      case STATE_ENDTAG:
        if (c0 == '>')
        {
          strTemp.assign(&acBuf[nBufIdxStart+1], nBufIdx-nBufIdxStart-1);
          if (strTemp.compare(myName) != 0)
             throw XMLException(sourcedescr + ": XML-Element " + getName() + " wurde nicht korrekt abgeschlossen: " + strTemp);
          fMachen = false;
        }
        break;
        
      default:
        break;
    }
    
    if (fMachen)
    {
      nBufIdx++;
      if (nBufIdx == nBufFill)
      {
        // data from nBufIdxStart to end must be moved to beginning of buffer
        int nFill = nBufFill-nBufIdxStart;
        if (nBufSize-nFill > nBufSize>>2)
        {
          nBufSize <<= 1;
          acBuf = (char*)realloc(acBuf, nBufSize);
        }
        if (nBufIdxStart)
        {
          memmove(&acBuf[0], &acBuf[nBufIdxStart], nFill);
          nBufIdx -= nBufIdxStart;
          nBufIdxStart = 0;
        }
        // read a block of data
        in.read(&acBuf[nFill], nBufSize-nFill);
        int nRead = in.gcount();
        nBufFill = nRead + nFill;
        if (nRead < 1)
          fMachen = false;
        else if (nBufIdx >= nBufFill)
        {
          throw XMLException("Internal error B: " + itoStr(nBufIdx, ' ', 1) + " >= " + itoStr(nBufFill, ' ', 1));
        }
      }
      else if (nBufIdx >= nBufFill)
      {
        throw XMLException("Internal error A: " + itoStr(nBufIdx, ' ', 1) + " >= " + itoStr(nBufFill, ' ', 1));
      }
    }
          
  }
  while (fMachen);
  
#if DEBUG == 1
  if (content != (std::string *) 0)
    printf("Content length=%i: %s", content->length(), content->c_str());
  else
    printf("No content\n");
#endif  
  
  // trailing whitespace vom content entfernen
  if (content != (std::string*) 0)
    while (!content->empty()
	   && isspace_utf8(content->at(content->length() - 1)) == true)
      content->resize(content->length()-1);
}

void SimpleXMLTransfer::removeChild(SimpleXMLTransfer* XMLPtr)
{
  bool fErr   = true;
  
  if (children.size())
  {
    for (unsigned int nIndex=0; nIndex<children.size(); nIndex++)
    {
      if (children[nIndex] == XMLPtr)
      {
        fErr = false;
        children.erase(children.begin() + nIndex);
        nIndex = children.size() + 1;
      }
    }    
  }
  
  if (fErr)
    throw
    XMLException
    ("void SimpleXMLTransfer::removeChild(...): no such child");
}

void SimpleXMLTransfer::removeChildAt(int nIndex)
{
  if (children.size() <= (unsigned int) nIndex)
    throw
    XMLException
    ("void SimpleXMLTransfer::removeChildAt(int nIndex): Index out of bounds in "
     + getName());
  
  children.erase(children.begin() + nIndex);
}

void SimpleXMLTransfer::replaceChild(SimpleXMLTransfer* XMLPtrOld,
                                     SimpleXMLTransfer* XMLPtrNew)
{
  bool fErr   = true;
  
  if (children.size())
  {
    for (unsigned int nIndex=0; nIndex<children.size(); nIndex++)
    {
      if (children[nIndex] == XMLPtrOld)
      {
        fErr = false;
        children[nIndex] = XMLPtrNew;
        nIndex = children.size() + 1;
      }
    }    
  }
  
  if (fErr)
    throw
    XMLException
    ("void SimpleXMLTransfer::replaceChild(...): no such child");
}

void SimpleXMLTransfer::setAttribute(std::string attribute, long int nValue)
{
  setAttribute(attribute, itoStr(nValue, ' ', 1));
}

void SimpleXMLTransfer::setAttribute(std::string attribute, std::string value)
{
  if (attribute.find('.') != std::string::npos)
  {
    int index;
    std::string childName;
    SimpleXMLTransfer* item = this;
    std::string::size_type pos;
    
    while ((pos = attribute.find('.')) != std::string::npos)
    {
      childName = attribute.substr(0, pos);
      attribute = attribute.substr(pos + 1);
      
      index = item->indexOfChild(childName);
      if (index < 0)
      {
        SimpleXMLTransfer* tmp = new SimpleXMLTransfer();
        if (tmp == 0)
          throw XMLException("setAttribute: error on new. out of memory?");

        tmp->setName(childName);
        item->addChild(tmp);
        item = tmp;
      }
      else
        item = item->getChildAt(index);
    }
    item->setAttribute(attribute, value);
  }
  else
  {
    if (indexOfAttribute(attribute) >= 0)
      throw XMLException("Attribute named " + attribute +
                         " already exists in " + getName());
    
    addAttribute(attribute, value);
  }
}

void SimpleXMLTransfer::setAttributeOverwrite(std::string attribute, long int nValue)
{
  setAttributeOverwrite(attribute, itoStr(nValue, ' ', 1));
}

void SimpleXMLTransfer::setAttributeOverwrite(std::string attribute, std::string value)
{
  if (attribute.find('.') != std::string::npos)
  {
    int index;
    std::string childName;
    SimpleXMLTransfer* item = this;
    std::string::size_type pos;
    
    while ((pos = attribute.find('.')) != std::string::npos)
    {
      childName = attribute.substr(0, pos);
      attribute = attribute.substr(pos + 1);
      
      index = item->indexOfChild(childName);
      if (index < 0)
      {
        SimpleXMLTransfer* tmp = new SimpleXMLTransfer();
        tmp->setName(childName);
        item->addChild(tmp);
        item = tmp;
      }
      else
        item = item->getChildAt(index);
    }
    
    item->setAttributeOverwrite(attribute, value);
  }
  else
  {
    int index = indexOfAttribute(attribute);
    if (index >= 0)
    {
      attrName.erase(attrName.begin() + index);
      attrVal.erase(attrVal.begin() + index);
    }
    
    addAttribute(attribute, value);
  }
}

void SimpleXMLTransfer::setContent(std::string newContent)
{
  if (content != (std::string *) 0)
    delete content;
  
  content = new std::string(newContent);
}

void SimpleXMLTransfer::setName(std::string name)
{
  myName = name;
}

void SimpleXMLTransfer::sortChildrenDouble(std::string attributeName)
{
  bool fChanged = true;
  int n;
  int size;
  double attr1;
  double attr2;
  SimpleXMLTransfer* tmp;
  
  size = children.size();
  
  while (fChanged)
  {
    fChanged = false;
    
    for (n = 0; n < size - 1; n++)
    {
      attr1 = children[n]->attributeAsDouble(attributeName, 0);
      attr2 = children[n + 1]->attributeAsDouble(attributeName, 0);
      
      if (attr1 > attr2)
      {
        tmp = children[n];
        children[n] = children[n + 1];
        children[n + 1] = tmp;
        fChanged = true;
      }
    }
    
    if (fChanged == true)
    {
      fChanged = false;
      
      for (n = size - 2; n >= 0; n--)
      {
        attr1 = children[n]->attributeAsDouble(attributeName, 0);
        attr2 = children[n + 1]->attributeAsDouble(attributeName, 0);
        
        if (attr1 > attr2)
        {
          tmp = children[n];
          children[n] = children[n + 1];
          children[n + 1] = tmp;
          fChanged = true;
        }
      }
    }
  }
}

void SimpleXMLTransfer::sortChildrenString(std::string attributeName)
{
  bool fChanged = true;
  int n;
  int size;
  std::string attr1;
  std::string attr2;
  SimpleXMLTransfer* tmp;
  
  size = children.size();
  
  while (fChanged)
  {
    fChanged = false;
    
    for (n = 0; n < size - 1; n++)
    {
      attr1 = children[n]->attribute(attributeName, "");
      attr2 = children[n + 1]->attribute(attributeName, "");
      
      if (attr1.compare(attr2) > 0)
      {
        tmp = children[n];
        children[n] = children[n + 1];
        children[n + 1] = tmp;
        fChanged = true;
      }
    }
    
    if (fChanged == true)
    {
      fChanged = false;
      
      for (n = size - 2; n >= 0; n--)
      {
        attr1 = children[n]->attribute(attributeName, "");
        attr2 = children[n + 1]->attribute(attributeName, "");
        
        if (attr1.compare(attr2) > 0)
        {
          tmp = children[n];
          children[n] = children[n + 1];
          children[n + 1] = tmp;
          fChanged = true;
        }
      }
    }
  }
}

void SimpleXMLTransfer::sortChildren(std::string attributeName, int (*pSortFun)(std::string a, std::string b))
{
  bool fChanged = true;
  int n;
  int size;
  std::string attr1;
  std::string attr2;
  SimpleXMLTransfer* tmp;
  
  size = children.size();
  
  while (fChanged)
  {
    fChanged = false;
    
    for (n = 0; n < size - 1; n++)
    {
      attr1 = children[n]->attribute(attributeName, "");
      attr2 = children[n + 1]->attribute(attributeName, "");
      
      if (pSortFun(attr1, attr2) > 0)
      {
        tmp = children[n];
        children[n] = children[n + 1];
        children[n + 1] = tmp;
        fChanged = true;
      }
    }
    
    if (fChanged == true)
    {
      fChanged = false;
      
      for (n = size - 2; n >= 0; n--)
      {
        attr1 = children[n]->attribute(attributeName, "");
        attr2 = children[n + 1]->attribute(attributeName, "");
        
        if (pSortFun(attr1, attr2) > 0)
        {
          tmp = children[n];
          children[n] = children[n + 1];
          children[n + 1] = tmp;
          fChanged = true;
        }
      }
    }
  }
}

void SimpleXMLTransfer::removeComments()
{
  comment.clear();
  commentPos.clear();

  int size = children.size();
  
  for (int n = 0; n < size; n++)
    children[n]->removeComments();
}

void SimpleXMLTransfer::insertChild(SimpleXMLTransfer* child, int idx)
{
  children.insert(children.begin()+idx, child);
}

int SimpleXMLTransfer::SaveToFile(std::string strFilename, int nIndent, int nIndentStep, int nLineBreak)
{
#ifdef WIN32_FLTK13
  FILE* fp = fl_fopen(strFilename.c_str(), "wb");
  if (fp == 0)
  {
    std::cerr << "error opening " << strFilename << " for writing\n";
    return(-1);
  }  
  oFILEstream out(fp);  
  if (out.fail())
  {
    std::cerr << "error after attaching filebuf to ostream: " << strFilename << "\n";
    return(-1);
  }
#else
  std::ofstream out;  
  out.open(strFilename.c_str());
  if (!out)
  {
    std::cerr << "error opening " << strFilename << " for writing\n";
    return(-1);
  }
#endif
  
  print(out, nIndent, nIndentStep, nLineBreak);

#ifdef WIN32_FLTK13
  //out.close();
#else
  out.close();
#endif

  int nRetCode = 0;
  if (out.fail())
  {
    std::cerr << "error after closing " << strFilename << ". bad=" << out.bad() << "\n";
    nRetCode = -1;
  }
  return(nRetCode);
}
