/*
 * Copyright (C) 2005-2009, 2012  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "filetools.h"

#include <fstream>
#include <iostream>
#include <sys/stat.h> 
#include <unistd.h> 
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
# include <direct.h>   // mkdir
# define WINVER 0x0500
# include <windows.h>  // DWORD
# include <Fcntl.h>    // flags for _open
#endif

#ifdef WIN32_FLTK13
# include <FL/fl_utf8.h>
# include "filestream.h"
#endif



std::string FileTools::createAbsPath(std::string base,
                                     std::string rel)
{
  if (rel.length() > 0 && rel[0] == '/')
    return(rel);
  else
  {
    std::string            path = base + rel;
    std::string            path1;
    std::string::size_type pos1;
    std::string::size_type pos2;
    
    // Reduce /./ to /
    while ((pos1 = path.find("/./")) != std::string::npos)
    {
      path1 = path.substr(0, pos1);
      path  = path.substr(pos1+2);
      path  = path1 + path;
    }
    
    // Reduce // to /
    while ((pos1 = path.find("//")) != std::string::npos)
      path  = path.replace(pos1, 2, "/");
    
    while ((pos1 = path.find("../")) != std::string::npos)
    {
      // std::cout << path << "\n";
      path1 = path.substr(0, pos1-1);
      path  = path.substr(pos1+2);
      
      if ((pos2 = path1.rfind('/')) != std::string::npos)
      {
        path1 = path1.substr(0, pos2);
        path  = path1 + path;
      }
    }
    
    return(path);
  }
}

std::string FileTools::createRelPath(std::string refPath,
                                     std::string absPath,
                                     bool        fLead)
{
  if (absPath.length() == 0)
    return "";
  
  //std::cout << refPath << " " << absPath << " " << fLead << "\n";
  
  std::string lead;
  if (fLead)
    lead = "./";
  else
    lead = "";
  
  if (absPath.find(refPath) == 0)
  {
    // Der einfachste Fall
    return(lead + absPath.substr(refPath.length()));
  }
  else
  {
    std::string            basis = refPath.substr(0, refPath.rfind('/')+1);
    std::string            hoch  = "";
    std::string::size_type pos;
    
    //std::cout << "basis:   " << basis << "\n";
    //std::cout << "absPath: " << absPath << "\n";
    
    // Den einen so lange hinten abschneiden, bis er in dem anderen vorkommt.
    pos = basis.length()-1;
    while (absPath.find(basis) == std::string::npos && (pos = basis.rfind('/', pos-1)) != std::string::npos && basis.length() > 1)
    {
      basis = basis.substr(0, pos+1);
      //std::cout << "basis: " << basis << "\n";
      hoch.append("../");
    }
    
    // 'Basis' ist jetzt die gemeinsame Basis, 'hoch' enthält den Weg
    // vom Referenzpfad zur gemeinsamen Basis.
          
    return(lead + hoch + absPath.substr(basis.length()));
  }  
}

bool FileTools::fileExists(std::string path)
{
#ifdef WIN32_FLTK13
  struct stat attribs;
  bool ret = FileTools::stat(path.c_str(), &attribs) == 0;
  return(ret);
#else
  std::ifstream testDat;
  
  // std::cout << "bool FileTools::fileExists(std::string path = " << path << ")\n";
  
  testDat.open((const char*)path.c_str());

  if (!testDat) 
    return(false);
  else 
  { 
    testDat.close();
    return(true);
  }  
#endif  
}

void FileTools::openAndCreateBackup(std::ofstream& out, std::string path)
{
  if (fileExists(path))
  {
    copy(path + "_backup", path);
  }
  
  out.open(path.c_str());
}

int FileTools::isFileUpToDate(std::string ref, std::string result)
{
  if (fileExists(result) == false)
  {
    // std::cout << "isFileUpToDate: result doesn't exist: " << result << "\n";
    return(-2);    
  }
  else
  {
    // take remaining data from file
    struct stat     AttribsRef;
    struct stat     AttribsRes;
    
    FileTools::stat(ref.c_str(),    &AttribsRef);
    FileTools::stat(result.c_str(), &AttribsRes);

    if (AttribsRes.st_mtime > AttribsRef.st_mtime)
    {
      // std::cout << "isFileUpToDate: ok: " << ref << " older than " << result << "\n";
      return(1);
    }
    else
    {
      // std::cout << "isFileUpToDate: too old: " << ref << " newer than " << result << "\n";
      return(-1);
    }
  }
}

int FileTools::makeSureDirExists(std::string dirname)
{
  int nRetCode = 0;
  
  // std::cout << "int FileTools::makeSureDirExists(" << dirname << ")\n";
  
  // I expect an absolute path beginning with '/'
  if (dirname.length() < 1)
    return(-1);
#ifndef WIN32
  else if (dirname[0] != '/')
    dirname = getWorkingDir() + "/" + dirname;
#endif

#ifdef WIN32_FLTK13
  fl_make_path(BackslashToSlash(dirname).c_str());
  return(0);
#endif    
  
  std::string::size_type posstart = 2;
  std::string::size_type pos;
  
  while (nRetCode == 0 && 
         posstart < dirname.length() && 
         (pos = dirname.find('/', posstart)) != std::string::npos)
  {
#ifdef WIN32
    mkdir(dirname.substr(0, pos).c_str());
#else
    if (mkdir(dirname.substr(0, pos).c_str(), 0777))
    {
      int nErrNr = errno;
      if (nErrNr != EEXIST)
      {
        // std::cout << "mkdir " << dirname.substr(0, pos) << "\n";
        std::cout << "Error on mkdir " << dirname.substr(0, pos) << ": " << strerror(nErrNr) << "\n";
        switch (nErrNr)
        {
          case EROFS:
            nRetCode = -200;
            break;
          default:
            nRetCode = -1;
            break;
        }
      }
    }
#endif    
    posstart = pos+1;
  }
  
  if (nRetCode == 0 && dirname[dirname.length()-1] != '/')
  {
#ifdef WIN32
    mkdir(dirname.c_str());
#else
    if (mkdir(dirname.c_str(), 0777))
    {
      int nErrNr = errno;
      if (errno != EEXIST)
      {
        std::cout << "Error on mkdir " << dirname.substr(0, pos) << ": " << strerror(nErrNr) << "\n";
        switch (nErrNr)
        {
          case EROFS:
            nRetCode = -200;
            break;
          default:
            nRetCode = -1;
            break;
        }
      }
    }
#endif
  }
          
  return(nRetCode);
}

std::string FileTools::getHomeDir(std::string appdirname)
{
  std::string tmp;
  
#ifdef WIN32
  tmp   = getenv("USERPROFILE");
  tmp  += "/";
  tmp  += appdirname;
#else
  tmp   = getenv("HOME");
  tmp  += "/.";
  tmp  += appdirname;
#endif      

  // Gibt es das Verzeichnis?
  struct stat     Attribs;              // ein Puffer für die Dateiattribute
  
  if (FileTools::stat(tmp.c_str(), &Attribs))
  {
    // Das Verzeichnis muss angelegt werden
#ifdef WIN32
    mkdir(tmp.c_str());
#else
    mkdir(tmp.c_str(), S_IRWXU);
#endif
    
    // Ist es jetzt vorhanden?
    if (FileTools::stat(tmp.c_str(), &Attribs))
    {
      // dann wird das aktuelle Verzeichnis benutzt
      tmp = "";
    }
    else
    {
      tmp += "/";
    }
  }
  else
  {
    tmp += "/";
  }
  
  // printf("DaLoMisc::getHomeDir() = %s\n", tmp.c_str());
  
  return(tmp);
}

int FileTools::rm(std::string filename)
{
#ifdef WIN32
# ifdef WIN32_FLTK13
  return(fl_unlink(filename.c_str()));
# else
  BOOL ret = DeleteFile(filename.c_str());
  if (ret)
    return(0);
  else
    return(-1);
# endif    
#else
  // linux: man 3 remove
  return(remove(filename.c_str()));
#endif
}

int FileTools::getSize(std::string filename)
{
  struct stat     Attribs;              // ein Puffer für die Dateiattribute
  
  if (FileTools::stat(filename.c_str(), &Attribs))
  {
    return(-1);
  }
  
  return(Attribs.st_size);
}

std::string FileTools::name(std::string pathAndName)
{
  std::string::size_type pos = pathAndName.rfind('/');
  
  if (pos != std::string::npos)
  {
    return(pathAndName.substr(pos+1));
  }
  else
    return(pathAndName);
}

std::string FileTools::path(std::string pathAndName)
{
  std::string::size_type pos = pathAndName.rfind('/');
  
  if (pos != std::string::npos)
  {
    return(pathAndName.substr(0, pos+1));
  }
  else
    return("");
}

std::string FileTools::removeSuffix(std::string name)
{
  std::string::size_type pos = name.find('.');
  
  if (pos != std::string::npos)
    return(name.substr(0, pos));
  else
    return(name);
}

std::string FileTools::getSuffix(std::string name)
{
  std::string::size_type pos = name.rfind('.');
  
  if (pos != std::string::npos)
    return(name.substr(pos));
  else
    return("");
}

int FileTools::copy(std::string dest, std::string src)
{
  if (dest[dest.length()-1] == '/')
    dest += name(src);
  
#ifdef WIN32_FLTK13
  FILE* fp_src = fl_fopen(src.c_str(), "rb");
#else
  FILE* fp_src = fopen(src.c_str(), "rb");
#endif
  if (fp_src == 0)
    return(-1);
  
#ifdef WIN32_FLTK13
  FILE* fp_dest = fl_fopen(dest.c_str(), "wb");
#else
  FILE* fp_dest = fopen(dest.c_str(), "wb");
#endif
  if (fp_dest == 0)
  {
    fclose(fp_src);
    return(-1);
  }    
        
  char buf[256];
  size_t n;
  int nRet = 0;
  
  while (feof(fp_src) == 0 && nRet == 0)
  {
    n = fread(buf, 1, 256, fp_src);
    if (fwrite(buf, 1, n, fp_dest) != n)
      nRet = -1;
  }
    
  fclose(fp_src);
  fclose(fp_dest);
  
  return(nRet);
}

int FileTools::move(std::string dest, std::string src)
{
#ifdef WIN32
  BOOL ret = MoveFileEx(src.c_str(), dest.c_str(), MOVEFILE_REPLACE_EXISTING | MOVEFILE_COPY_ALLOWED );
  if (ret)
    return(0);
  else
    return(-1);
#else   
  if (dest[dest.length()-1] == '/')
    dest += name(src);
  
  std::string call  = "mv \"";
  call += src;
  call += "\" \"" + dest + "\"";
  
  return(
         system(call.c_str())
         );
#endif
}

int FileTools::symlink(std::string dest, std::string src)
{
#ifdef WIN32
  return(-1);
#else
  return(::symlink(src.c_str(), dest.c_str()));
#endif
}

std::string FileTools::stripAbsFromPath(std::string in)
{
#ifdef WIN32
  if (in.length() > 1 && in[1] == ':')
    in = in.substr(2);
#endif
  while (in.length() && in[0] == '/')
    in = in.substr(1);
  
  return(in);
}

std::string FileTools::cleanString(std::string pathAndName, std::string allowed)
{
  int len = pathAndName.length();
  
  for (int cnt=0; cnt<len; cnt++)
  {
    if ((isalnum(pathAndName[cnt]) == 0 || (pathAndName[cnt] & 0x80) != 0) &&
        pathAndName[cnt] != '_' &&
        allowed.find(pathAndName[cnt]) == std::string::npos)
    {
      pathAndName[cnt] = '_';
    }        
  }
  
  return(pathAndName);
}

bool FileTools::isDir(std::string name)
{
  struct stat     AttribsRef;
    
  FileTools::stat(name.c_str(), &AttribsRef);
   
  return(S_ISDIR(AttribsRef.st_mode) != 0);
}

bool FileTools::isLink(std::string name)
{
#ifdef WIN32
  return(false);
#else
  struct stat     AttribsRef;
    
  lstat(name.c_str(),    &AttribsRef);
   
  return(S_ISLNK(AttribsRef.st_mode) != 0);
#endif
}

bool FileTools::isWriteable(std::string name)
{
  return(true); // todo
}

time_t FileTools::Get_mtime(std::string path)
{
  if (fileExists(path))
  {
    struct stat AttribsRef;
    FileTools::stat(path.c_str(), &AttribsRef);
  
    return(AttribsRef.st_mtime);
  }
  else
  {
    return(0);
  }
}

/**
 * Wohin zeigt ein Link?
 */
std::string FileTools::getLinkTarget(std::string path)
{
#ifdef WIN32
   return("");
#else
  char*        buf;
  size_t       bufsize = 200;
  std::string::size_type pos;
  
  buf = (char*)calloc(bufsize, 1);

  while (readlink(path.c_str(), buf, bufsize) == -1)
  {
    bufsize <<= 1;
    // std::cout << bufsize << "\n";
    free(buf);
    buf = (char*)calloc(bufsize, 1);
    if (bufsize > 4000)
      return("");
  }
  if (buf == (char*)-1)
  {
    std::cerr << "Error on calloc:\n  " << path << "\n";
    exit(-1);
  }
  
  // Der Inhalt steht jetzt in buf, aber nicht NUL-terminated.
  // Daher muss man calloc benutzen...
  path = buf;
  
  free(buf);
  
  // Evtl. vorkommende "/./" entfernen
  while ((pos = path.find("/./")) != std::string::npos)
  {
    std::string path1 = path.substr(0, pos);
    
    path  = path.substr(pos+2);
    path  = path1 + path;
  }
    
  // Evtl. vorkommende "./" zu Beginn entfernen
  if (path.find("./") == 0)
    path  = path.substr(2);
    
  return(path);
#endif
}

std::string FileTools::BackslashToSlash(std::string in)
{
  std::string::size_type pos=0;
  while ((pos = in.find('\\', pos)) != std::string::npos)
  {
    in[pos] = '/';
  }
  return(in);
}

int FileTools::stat(std::string strName, struct stat* attribs)
{
#ifdef WIN32_FLTK13
  // convert utf8 to wide character before calling wstat
//  int len = strName.length();
//  int outlen = len*2+1;
//  wchar_t* buf = new wchar_t[outlen];
//  fl_utf8towc(strName.c_str(), len, buf, outlen);
//  return(_wstat(buf, attribs));
//  delete[] buf;
  return(fl_stat(strName.c_str(), attribs));
#else
  return(::stat(strName.c_str(), attribs));
#endif  
} 


#if (TEMP_TEXTFILE_VARIANT == 0)
TemporaryTextFile::TemporaryTextFile(std::string name_template)
{
  name_template += "XXXXXX";
  
  char* buf = (char*)malloc(name_template.length()+1);
  strcpy(buf, name_template.c_str());
  
  filedescr = mkstemp(buf);
  if (filedescr != -1)
  {
    fchmod(filedescr, S_IWUSR | S_IRUSR);
    FilenameAndPath = buf;
    fOK = true;
  }
  else
  {
    fOK = false;
  }        
  free(buf);
}
#endif

#if (TEMP_TEXTFILE_VARIANT == 1)
TemporaryTextFile::TemporaryTextFile(std::string name_template)
{
  char         buf[10];
  const int    cMaxCount = 1000000;
  int          cnt       = 0;
  std::string  name;
  
  do 
  {
    sprintf(buf, "%i", cnt);
    name = name_template + buf;
    cnt++;
  } while (FileTools::fileExists(name) && cnt < cMaxCount);
  
  if (cnt == cMaxCount)
    fOK = false;
  else
  {
    outfile.open(name.c_str());
    if (outfile)
    {
      FilenameAndPath = name;      
      fOK = true;
    }
    else
      fOK = false;
  }  
}
#endif

#if (TEMP_TEXTFILE_VARIANT == 2)
TemporaryTextFile::TemporaryTextFile()
{
  tmpfiledescr = tmpfile();
  if (tmpfiledescr != NULL)
  {
    // big trouble: what is the filename?
    FilenameAndPath = "";    
    fOK = true;    
  }
  else
  {
    fOK = false;
  }
}
#endif

void TemporaryTextFile::write(std::string line)
{
#if (TEMP_TEXTFILE_VARIANT == 0)
  ::write(filedescr, line.c_str(), line.length());
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 1)
  outfile << line.c_str();
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 2)
  std::string  tmpstr;
  unsigned int uLen;
  
  // Work around potential security hole (see man fprintf):
  uLen = tmpstr.length();
  for (unsigned int uCnt=0; uCnt<uLen; uCnt++)
    if (tmpstr[uCnt] == '%')
      tmpstr[uCnt] = ' ';
  fprintf(tmpfiledescr, "%s", tmpstr.c_str());
#endif
}

void TemporaryTextFile::close()
{
#if (TEMP_TEXTFILE_VARIANT == 0)
  fsync(filedescr);
  ::close(filedescr);
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 1)
  outfile.close();
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 2)
  // File would be deleted on close. Something like fsync?
#endif
}

std::string TemporaryTextFile::GetGlobalTempDirPath()
{
#ifdef WIN32
# ifdef WIN32_FLTK13
   DWORD len = ::GetTempPath(0, "");
   if (len == 0)
   {
     return("");
   }   
   else
   {
     int size = len+1;
     WCHAR* tmpbuf = (WCHAR*)malloc(size*sizeof(WCHAR));
     ::GetTempPathW((DWORD)(size), tmpbuf);
     DWORD longbuflen = GetLongPathNameW(tmpbuf, 0, 0);
     WCHAR* longbuf = (WCHAR*)malloc(longbuflen*sizeof(WCHAR));
     GetLongPathNameW(tmpbuf, longbuf, longbuflen);
     free(tmpbuf);
     
     char* utfbuf = (char*)malloc(longbuflen*6);
     fl_utf8fromwc(utfbuf, longbuflen*6, longbuf, longbuflen);
     free(longbuf);
     
     std::string tmp;
     tmp.assign(utfbuf);
     free(utfbuf);
     
     return(tmp);
   }  

# else
   // alternatively the enviroment variable TEMP can be used
   DWORD len = ::GetTempPath(0, "");
   if (len == 0)
   {
     return("");
   }   
   else
   {
     int size = len+1;
     char* tmpbuf = (char*)malloc(size);
     std::string tmp;
     len = ::GetTempPath((DWORD)(size), tmpbuf);
     tmp = tmpbuf;
     free(tmpbuf);
     return(tmp);
   }  
# endif
#else
  return("/tmp/");
#endif
}
int TemporaryTextFile::GetTempFD(std::string& path)
{
#ifdef WIN32
   path = GetGlobalTempDirPath().c_str();
   char buffer[MAX_PATH];   
   GetTempFileName(path.c_str(), "bla", 0, buffer);
   path = buffer;
   return(_open(path.c_str(), _O_BINARY | _O_CREAT | _O_RDWR, _S_IREAD | _S_IWRITE));
#else
   char tempfile[21];
   strcpy(tempfile, "/tmp/tempfXXXXXX");
   int tempfd = mkstemp(tempfile);
   fchmod(tempfd, S_IWUSR | S_IRUSR);
   path = tempfile;
   return(tempfd);
#endif
}

std::string FileTools::getWorkingDir()
{
  std::string tmp = "";
  
  size_t size = 100;
     
  while (1)
  {
    char *buffer = (char *) malloc (size);
    if (getcwd (buffer, size) == buffer)
    {
      tmp = buffer;
      free(buffer);
      return(tmp);
    }
    free (buffer);
    if (errno != ERANGE)
      return("");
    size *= 2;
  }
}

